# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
import numpy as np
from typing import List, Tuple, Any, Callable

from model_compression_toolkit.common import FrameworkInfo, Graph, BaseNode
from model_compression_toolkit.common.constants import THRESHOLD, SIGNED, SHIFT_NEGATIVE_NON_LINEAR_NUM_BITS
from model_compression_toolkit.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.common.hardware_representation import QuantizationMethod
from model_compression_toolkit.common.quantization.set_node_quantization_config import create_node_activation_qc, \
    set_quantization_configs_to_node
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig
from model_compression_toolkit.common.quantization.quantization_params_generation.qparams_activations_computation \
    import get_activations_qparams
from model_compression_toolkit.keras.constants import PADDING

"""
This substitution aims to solve an issue of activation with negative outputs where
the portion of the negative range is relatively small. In a symmetric quantization this causes 
of bit loosing as the entire negative quantization range does not contain
any values. To solve it, we shift the output of the activation by the minimal output value (quantized) such
that all values after the shifting are positive. To correct the impact of such shifting, a correction
to the next linear node is computed and added to its bias term.
If the linear node pads the input tensor with zeros, we modify the padded value as well.  
"""


def op2d_bias_correction(op2d_node: BaseNode,
                         shift_to_correct: float,
                         fw_info: FrameworkInfo,
                         bias_str: str,
                         bias_flag_str: str):
    """
    Compute the correction term to add to the op2d node's bias
    to correct the error occurs from adding an Add node (shifting).

    Args:
        op2d_node: Node to compute its bias correction term.
        shift_to_correct: Value that was used to shift the output tensor of
        the non-linear node.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
        bias_str:
        bias_flag_str: The framework specific attribute name of the bias flag.
    """

    bias = op2d_node.get_weights_by_keys(bias_str)
    if bias is None:
        bias = 0.0
        op2d_node.framework_attr[bias_flag_str] = True

    # Each node adds a different noise due to the shifting. It depends on the
    # dimensions of the kernel, thus the correction term is a function of
    # the layer type.
    kernel = op2d_node.get_weights_by_keys(fw_info.kernel_ops_attributes_mapping.get(op2d_node.type)[0])
    if kernel is not None:
        output_channel_index, input_channel_index = fw_info.kernel_channels_mapping.get(op2d_node.type)
        axis_not_output_channel = list(range(len(kernel.shape)))
        axis_not_output_channel.remove(output_channel_index)

        # special case of depthwise_conv2d in tensorflow, where we have a depth multiplier for the filters
        if output_channel_index == input_channel_index:
            axis_not_output_channel.remove(3) # 3 is the depth multiplier index

        bias_correction = shift_to_correct * np.sum(kernel, axis=tuple(axis_not_output_channel))
        op2d_node.set_weights_by_keys(bias_str, bias - bias_correction.flatten())
    else:
        raise NotImplementedError


def insert_node_between_two_nodes(graph: Graph,
                                  node_to_insert: BaseNode,
                                  first_node: BaseNode,
                                  last_node: BaseNode):
    """
    Insert a new node in a graph between two nodes.

    Args:
        graph: Graph to add the new node to.
        node_to_insert: Node to add.
        first_node: Node to insert the new node after it.
        last_node: Node to insert the new node before it.

    """

    graph.add_node(node_to_insert)
    e_attr = graph.get_edge_data(first_node, last_node)
    assert len(list(e_attr.values())) == 1
    e_attr = list(e_attr.values())[0]
    graph.add_edge(first_node, node_to_insert, **e_attr)
    graph.add_edge(node_to_insert, last_node, **e_attr)
    graph.remove_edge(first_node, last_node)


def insert_node_after_node(graph: Graph,
                           node_to_insert: BaseNode,
                           first_node: BaseNode):
    """
    Insert a new node to a graph after an existing node in the graph.
    Check before insertion that the node (that we add the new node after) has
    only a single outgoing edge, so such an insertion is possible. If it is not the
    case, an exception is thrown.

    Args:
        graph: Graph to add the new node to.
        node_to_insert: Node to add.
        first_node: Node to insert the new node after it.

    """

    last_nodes = graph.get_next_nodes(first_node)
    if len(last_nodes) != 1:
        raise Exception('Can only insert if there is only one input')
    last_node = last_nodes[0]
    insert_node_between_two_nodes(graph, node_to_insert, first_node, last_node)


def insert_node_before_node(graph: Graph,
                            node_to_insert: BaseNode,
                            last_node: BaseNode):
    """
    Insert a new node to a graph before an existing node in the graph.
    Check before insertion that the node (that we add the new node before) has
    only a single incoming edge, so such an insertion is possible. If it is not the
    case, an exception is thrown.

    Args:
        graph: Graph to add the new node to.
        node_to_insert: Node to add.
        last_node: Node to insert the new node after it.

    """
    first_nodes = graph.get_prev_nodes(last_node)
    if len(first_nodes) != 1:
        raise Exception('Can only insert if there is only one input')
    first_node = first_nodes[0]
    insert_node_between_two_nodes(graph, node_to_insert, first_node, last_node)


def remove_node_between_two_nodes(graph: Graph,
                                  node_to_remove: BaseNode,
                                  first_node: BaseNode,
                                  last_node: BaseNode):
    """
    Remove a node from a graph and connect its previous node to
    its next node after the removal.

    Args:
        graph: Graph to modify.
        node_to_remove: Node to remove from the graph.
        first_node: Previous node to the node to be removed.
        last_node: Next node to the node to be removed.

    """

    e_attr = graph.get_edge_data(first_node, node_to_remove)
    assert len(list(e_attr.values())) == 1
    e_attr = list(e_attr.values())[0]
    graph.add_edge(first_node, last_node, **e_attr)

    graph.remove_edge(first_node, node_to_remove)
    graph.remove_edge(node_to_remove, last_node)
    graph.remove_node(node_to_remove)


def shift_negative_function(graph: Graph,
                            qc: QuantizationConfig,
                            non_linear_node: BaseNode,
                            op2d_node: BaseNode,
                            fw_info: FrameworkInfo,
                            create_add_node: Callable,
                            get_padding_values: Callable,
                            create_pad_node: Callable,
                            padding_str: str,
                            bias_str: str,
                            bias_flag_str: str,
                            zero_padding_node: BaseNode = None,
                            bypass_nodes: List = None,
                            ) -> Graph:
    """
    Shift the output of a non-linear activation by its minimal output value (quantized) such
    that all values after the shifting are positive.
    The shifting happens only if the ratio between the shifting value and the threshold is small enough
    (the threshold to activate the shifting and correction is in the passed QuantizationConfig, qc).
    To correct the impact of such shifting, a correction to the next linear node is computed and
    added to its bias term.
    If the linear node pads the input tensor with zeros, we modify the padded value as well.

    Args:
        graph: Graph to apply the shifting and correction.
        qc: Quantization configuration to build the substitutions list according to.
        non_linear_node: Non-linear node with negative values to shift.
        op2d_node: Linear node to correct its bias to overcome the expected error due to
        the shifting.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
        groups of layers by how they should be quantized, etc.)
        create_add_node: Function to create an add node.
        get_padding_values: Function to compute the op2d node's padding values
        create_pad_node: Function to create an pad node.
        padding_str: The framework specific attribute name of the padding.
        bias_str: The framework specific attribute name of the bias.
        bias_flag_str: The framework specific attribute name of the bias flag.
        zero_padding_node: ZeroPadding2D node that may be in the graph before the linear layer.

    Returns:
        Graph after applying the shifting and correction.
    """

    min_to_correct, max_value2compare = graph.get_out_stats_collector(non_linear_node).get_min_max_values()

    if not non_linear_node.is_all_activation_candidates_equal():
        raise Exception("Shift negative correction is not supported for more than one activation quantization "
                        "configuration candidate")

    # all candidates have same activation config, so taking the first candidate for calculations
    non_linear_node_cfg_candidate = non_linear_node.candidates_quantization_cfg[0].activation_quantization_cfg

    # get the non-linear activation threshold
    activation_threshold = non_linear_node_cfg_candidate.activation_quantization_params.get(THRESHOLD)

    negative_rate = np.abs(min_to_correct) / activation_threshold

    enable_sub = negative_rate <= non_linear_node_cfg_candidate.shift_negative_ratio
    if min_to_correct >= 0 or not enable_sub:
        return graph

    # Calculate the shifting value by checking the quantized points of the shifted activation and
    # taking the minimal quantized point that is still positive.
    q_points = np.linspace(0, activation_threshold - activation_threshold / (
            2 ** non_linear_node_cfg_candidate.activation_n_bits),
                           2 ** non_linear_node_cfg_candidate.activation_n_bits).astype(
        'float32')  # Change to type float32 to support tensorflow dtypes

    delta = q_points + min_to_correct
    delta[delta < 0] = np.inf
    shift_value = q_points[np.argmin(delta)]

    if zero_padding_node is not None:
        # Remove zero padding layer and save padding values for creating new pad layer
        padding = zero_padding_node.framework_attr.get(padding_str)
        pad_top, pad_btm, pad_left, pad_right = padding[0][0], padding[0][1], padding[1][0], padding[1][1]
        remove_node_between_two_nodes(graph,
                                      node_to_remove=zero_padding_node,
                                      first_node=non_linear_node,
                                      last_node=op2d_node)

    else:
        padding, padding_values = get_padding_values(op2d_node)
        if padding_values is not None:
            pad_top, pad_btm, pad_left, pad_right = padding_values

    # Insert Add node between non linear node to op2d, and fix op2d bias
    add_node = create_add_node(shift_value,
                               non_linear_node.name,
                               non_linear_node.input_shape)
    insert_node_after_node(graph,
                           node_to_insert=add_node,
                           first_node=non_linear_node)
    op2d_bias_correction(op2d_node,
                         shift_value,
                         fw_info,
                         bias_str,
                         bias_flag_str)

    # Use non linear statistics to create statistics for the Add node according to the shifting
    nl_stats_collector = graph.get_out_stats_collector(non_linear_node)

    add_node_stats_collector = copy.copy(nl_stats_collector)
    graph.set_out_stats_collector_to_node(add_node, add_node_stats_collector)
    graph.shift_stats_collector(add_node, np.array(shift_value))

    if padding is not None:
        pad_node = create_pad_node(op2d_node.name,
                                   add_node.name,
                                   shift_value,
                                   add_node.output_shape,
                                   pad_top, pad_btm, pad_left, pad_right)

        # Set quantization configuration to node, even though we do not quantize it:
        set_quantization_configs_to_node(fw_info=fw_info,
                                         node=pad_node,
                                         quant_config=qc,
                                         fw_hw_model=graph.fw_hw_model)

        for candidate_qc in pad_node.candidates_quantization_cfg:
            candidate_qc.weights_quantization_cfg.enable_weights_quantization = False
            candidate_qc.activation_quantization_cfg.enable_activation_quantization = False

        # Insert a pad node between the add node to the op2d, and create statistics for the pad node
        insert_node_before_node(graph,
                                node_to_insert=pad_node,
                                last_node=op2d_node)

        graph.set_out_stats_collector_to_node(pad_node,
                                              add_node_stats_collector)  # We ignore the padding effect on statistics

        op2d_node.input_shape = pad_node.output_shape

    set_quantization_configs_to_node(fw_info=fw_info,
                                     node=add_node,
                                     quant_config=qc,
                                     fw_hw_model=graph.fw_hw_model)

    for candidate_qc in add_node.candidates_quantization_cfg:
        candidate_qc.activation_quantization_cfg.activation_n_bits = \
            non_linear_node_cfg_candidate.activation_n_bits
    # The non-linear node's output should be float, so we approximate it by using 16bits quantization.
    for candidate_qc in non_linear_node.candidates_quantization_cfg:
        candidate_qc.activation_quantization_cfg.activation_n_bits = SHIFT_NEGATIVE_NON_LINEAR_NUM_BITS

    # A bypass node that has its own activation (e.g. GlobalAvgPool2D) can set it to unsigned
    if bypass_nodes:
        for bypass_node in bypass_nodes:
            for bypass_candidate_qc in bypass_node.candidates_quantization_cfg:
                if bypass_candidate_qc.activation_quantization_cfg:
                    bypass_candidate_qc.activation_quantization_cfg.activation_quantization_params[SIGNED] = False
                    graph.shift_stats_collector(bypass_node, np.array(shift_value))

    for candidate_qc in add_node.candidates_quantization_cfg:
        candidate_qc.weights_quantization_cfg.enable_weights_quantization = False

        candidate_qc.activation_quantization_cfg = create_node_activation_qc(qc,
                                                                             fw_info,
                                                                             graph.fw_hw_model.get_default_op_qc())

        candidate_qc.activation_quantization_cfg.set_activation_quantization_param({THRESHOLD: activation_threshold,
                                                                                    SIGNED: False})

    if non_linear_node_cfg_candidate.shift_negative_threshold_recalculation:
        activation_param = get_activations_qparams(activation_quant_cfg=non_linear_node_cfg_candidate,
                                                   nodes_prior_info=non_linear_node.prior_info,
                                                   out_stats_container=graph.get_out_stats_collector(non_linear_node))

        assert activation_param.get(SIGNED) == False
        for candidate_qc in non_linear_node.candidates_quantization_cfg:
            candidate_qc.activation_quantization_cfg.set_activation_quantization_param(activation_param)

    return graph


def get_next_nodes_to_correct(n: BaseNode,
                              graph: Graph,
                              linear_node_types: NodeOperationMatcher,
                              bypass_node_types: NodeOperationMatcher,
                              pad_node_types: NodeOperationMatcher,
                              is_padding_node_and_node_has_padding: Callable,
                              pad_node_to_consider: BaseNode = None,
                              bypass_nodes: List = None) -> Tuple[Any, Any, Any]:
    """
    Search for the next linear node of a given node. Go over
    the next nodes of the node and recursively search for a linear node.

    Args:
        n: Node to search for its next linear node.
        graph: Graph the node is in.
        linear_node_types: Types of linear nodes to consider.
        bypass_node_types: Types of nodes for bypassing to consider.
        pad_node_types: Types of padding nodes to consider.
        is_padding_node_and_node_has_padding: Function to check whether a padding node exists and
         the next node is a linear node with padding.
        pad_node_to_consider: Pad node between the non-linear and linear nodes to consider when
        correcting the expected shift.
        bypass_nodes: a list of bypass nodes found while running this function

    Returns:
        The linear node (if found), a padding node (if found) and a list of bypass nodes (if any), or Nones if it
        were not found or there are multiple outgoing edges to one of nodes during the search (which means, the
        substitution can not be applied).
    """

    next_nodes = graph.get_next_nodes(n)

    if len(next_nodes) != 1:
        return None, None, None

    next_node = next_nodes[0]

    if linear_node_types.apply(next_node):
        # Correction is not supported when there are both padding node and a linear node with padding.
        if is_padding_node_and_node_has_padding(pad_node_to_consider, next_node):
            return None, None, None
        return next_node, pad_node_to_consider, bypass_nodes

    if bypass_node_types.apply(next_node):
        if bypass_nodes:
            bypass_nodes.append(next_node)
        else:
            bypass_nodes = [next_node]
        return get_next_nodes_to_correct(next_node,
                                         graph,
                                         linear_node_types,
                                         bypass_node_types,
                                         pad_node_types,
                                         is_padding_node_and_node_has_padding,
                                         pad_node_to_consider,
                                         bypass_nodes=bypass_nodes)

    if pad_node_types.apply(next_node):
        # Correction is not supported when there are more than one padding node between the non-linear node and the
        # linear node.
        if pad_node_to_consider is None:
            return get_next_nodes_to_correct(next_node,
                                             graph,
                                             linear_node_types,
                                             bypass_node_types,
                                             pad_node_types,
                                             is_padding_node_and_node_has_padding,
                                             next_node,
                                             bypass_nodes=bypass_nodes)

    return None, None, None  # If none of the above were found, it means the correction can not be applied


def apply_shift_negative_correction(graph: Graph,
                                    quant_config: QuantizationConfig,
                                    fw_info: FrameworkInfo,
                                    snc_node_types: NodeOperationMatcher,
                                    linear_node_types: NodeOperationMatcher,
                                    bypass_node_types: NodeOperationMatcher,
                                    pad_node_types: NodeOperationMatcher,
                                    create_add_node: Callable,
                                    get_padding_values: Callable,
                                    create_pad_node: Callable,
                                    is_padding_node_and_node_has_padding: Callable,
                                    padding_str: str,
                                    bias_str: str,
                                    bias_flag_str: str) -> Graph:
    """
    Apply the substitution even if the linear node is not immediately after
    the non-linear node, but there are intermediate nodes

    Args:
        graph: Graph to apply the substitution on.
        quant_config: Quantization configuration to build the substitutions list according to.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
        groups of layers by how they should be quantized, etc.)
        snc_node_types: Types of activation nodes with negative outputs to consider.
        linear_node_types: Types of linear nodes to consider.
        bypass_node_types: Types of nodes for bypassing to consider.
        pad_node_types: Types of padding nodes to consider.
        create_add_node: Function to create an add node.
        get_padding_values: Function to compute the op2d node's padding values.
        create_pad_node: Function to create an pad node.
        is_padding_node_and_node_has_padding: Function to check whether a padding node exists and
         the next node is a linear node with padding.
        padding_str: The framework specific attribute name of the padding.
        bias_str: The framework specific attribute name of the bias.
        bias_flag_str: The framework specific attribute name of the bias flag.
    Returns:
        Graph after applying shift negative on selected activations.
    """
    # Skip substitution if QuantizationMethod is uniform.
    op_qc = graph.fw_hw_model.get_default_op_qc()
    if op_qc.activation_quantization_method is QuantizationMethod.UNIFORM:
        return graph

    nodes = list(graph.nodes())
    for n in nodes:
        if snc_node_types.apply(n):
            linear_node, pad_node, bypass_nodes = get_next_nodes_to_correct(n,
                                                                            graph,
                                                                            linear_node_types,
                                                                            bypass_node_types,
                                                                            pad_node_types,
                                                                            is_padding_node_and_node_has_padding
                                                                            )
            if linear_node is not None:
                graph = shift_negative_function(graph,
                                                quant_config,
                                                n,
                                                linear_node,
                                                fw_info,
                                                create_add_node,
                                                get_padding_values,
                                                create_pad_node,
                                                padding_str,
                                                bias_str,
                                                bias_flag_str,
                                                zero_padding_node=pad_node,
                                                bypass_nodes=bypass_nodes)
    return graph



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import numpy as np
from typing import List

from model_compression_toolkit import FrameworkInfo, common
from model_compression_toolkit.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.logger import Logger
from model_compression_toolkit.common.model_builder_mode import ModelBuilderMode


class ModelCollector(object):
    """
    Build a Keras model from a graph for statistics collection purposes.
    The ModelCollector builds a float model that its outputs are all layers outputs, so after
    inferring using this model, statistics of output layers can be gathered and be used
    for thresholds calculations.
    """

    def __init__(self, graph: Graph, fw_impl: FrameworkImplementation, fw_info: FrameworkInfo):
        """
        Build a Keras model from the passed graph, and set the model's
        outputs to be all layers' outputs.

        Args:
            graph: Graph to build a model from it.
            fw_impl: FrameworkImplementation object with a specific framework methods implementation.

        """

        self.graph = graph
        self.fw_impl = fw_impl
        self.fw_info = fw_info

        node2fetch = []  # List of graph nodes, the model should output their outputs.
        stats_containers_list = []  # List of output statistics containers of nodes ordered
        # the same as node2fetch so statistics of outputs can be gathered for the correct statistics container.

        for n in self.graph.nodes():
            out_stats_container = self.graph.get_out_stats_collector(n)
            if isinstance(out_stats_container, list):  # If layer has multiple outputs
                # Append nodes to output and track their statistics only if
                # they actually collect statistics.
                if len([x for x in out_stats_container if not isinstance(x, common.NoStatsCollector)]) > 0:
                    mark2fetch = True
                    for sc in out_stats_container:
                        # Output only if statistics should be gathered
                        if sc.require_collection() and mark2fetch:
                            mark2fetch = False
                            node2fetch.append(n)  # Append node several times (as number of outputs it has)
                    stats_containers_list.append(out_stats_container)

            else:  # A single output
                if out_stats_container.require_collection():
                    node2fetch.append(n)
                    stats_containers_list.append(out_stats_container)

        self.stats_containers_list = stats_containers_list

        # Build a float model and output all layers' outputs
        # (that should be collected) as the model's outputs
        self.model, _ = self.fw_impl.model_builder(self.graph,
                                                   mode=ModelBuilderMode.FLOAT,
                                                   append2output=node2fetch,
                                                   fw_info=self.fw_info)

    def infer(self, inputs_list: List[np.ndarray]):
        """
        Pass inputs through the model of the ModelCollector,
        and update statistics in all statistics containers the ModelCollector holds.

        Args:
            inputs_list: Inputs for the model inferring.

        """

        # TODO: Thinking about delegating collections to framework
        # TODO: migrate datasets to framework datasets
        tensor_data = self.fw_impl.run_model_inference(self.model, inputs_list)
        for td, sc in zip(tensor_data, self.stats_containers_list):
            if isinstance(sc, (list, tuple)):
                if not isinstance(td, (list, tuple)):
                    Logger.exception(
                        '"tensor_data" must be a list or a tuple if the model tensor_list is a list or a tuple')
                if len(sc) != len(td):
                    Logger.exception('"tensor_data" and the model tensor_list must be of the same length')
                for tdi, sci in zip(td, sc):
                    sci.update_statistics(self.fw_impl.to_numpy(tdi))
            else:
                sc.update_statistics(self.fw_impl.to_numpy(td))



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Any, Dict

from model_compression_toolkit.common.graph.base_node import BaseNode
from model_compression_toolkit.common.hardware_representation.hardware2framework.attribute_filter import AttributeFilter


class LayerFilterParams:
    """
    Wrap a layer with filters to filter framework's layers by their attributes.
    """

    def __init__(self, layer,
                 *conditions,
                 **kwargs):
        """

        Args:
            layer: Layer to match when filtering.
            *conditions (AttributeFilter): List of conditions to satisfy.
            **kwargs: Keyword arguments to filter layers according to.
        """
        self.layer = layer
        self.conditions = conditions
        self.kwargs = kwargs
        self.__name__ = self.create_name()

    def __hash__(self):
        """

        Returns: Hash code for the the LayerFilterParams. Used to check if a LayerFilterParams
        is mapped to multiple OperatorsSet.

        """
        return hash(self.__name__)  # TODO: reuven: will hash differently if conditions/kwargs are in different order. fix it

    def __eq__(self, other: Any) -> bool:
        """
        Check if an object is equal to the LayerFilterParams.

        Args:
            other: Object to check.

        Returns:
            Whether an object is equal to the LayerFilterParams.
        """

        if not isinstance(other, LayerFilterParams):
            return False

        # Check equality of conditions
        for self_c, other_c in zip(self.conditions, other.conditions):
            if self_c != other_c:
                return False

        # Check key-value arguments equality
        for k, v in self.kwargs.items():
            if k not in other.kwargs:
                return False
            else:
                if other.kwargs.get(k) != v:
                    return False
        return True

    def create_name(self) -> str:
        """

        Returns: Name of the LayerFilterParams. The name is composed of the layer type,
        conditions and keyword arguments. Used for display and hashing.

        """
        params = [f'{k}={v}' for k,v in self.kwargs.items()]
        params.extend([str(c) for c in self.conditions])
        params_str = ', '.join(params)
        return f'{self.layer.__name__}({params_str})'

    def match(self,
              node: BaseNode) -> bool:
        """
        Check if a node matches the layer, conditions and keyword-arguments of
        the LayerFilterParams.

        Args:
            node: Node to check if matches to the LayerFilterParams properties.

        Returns:
            Whether the node matches to the LayerFilterParams properties.
        """
        # Check the node has the same type as the layer in LayerFilterParams
        if self.layer != node.type:
            return False

        # Get attributes from node to filter
        layer_config = node.framework_attr
        if hasattr(node, "op_call_kwargs"):
            layer_config.update(node.op_call_kwargs)

        for attr, value in self.kwargs.items():
            if layer_config.get(attr) != value:
                return False

        for c in self.conditions:
            if not c.match(layer_config):
                return False

        return True



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit import hardware_representation as hw_model


def get_qnnpack_model():
    # Create a quantization config. A quantization configuration defines how an operator
    # should be quantized on the modeled hardware.
    # For qnnpack backend, Pytorch uses a QConfig with torch.per_tensor_affine for
    # activations quantization and a torch.per_tensor_symmetric quantization scheme
    # for weights quantization (https://pytorch.org/docs/stable/quantization.html#natively-supported-backends):
    eight_bits = hw_model.OpQuantizationConfig(
        activation_quantization_method=hw_model.QuantizationMethod.UNIFORM,
        weights_quantization_method=hw_model.QuantizationMethod.SYMMETRIC,
        activation_n_bits=8,
        weights_n_bits=8,
        weights_per_channel_threshold=False,
        enable_weights_quantization=True,
        enable_activation_quantization=True,
        quantization_preserving=False,
        fixed_scale=None,
        fixed_zero_point=None,
        weights_multiplier_nbits=None
    )

    # Create a QuantizationConfigOptions, which defines a set
    # of possible configurations to consider when quantizing a set of operations (in mixed-precision, for example).
    # If the QuantizationConfigOptions contains only one configuration,
    # this configuration will be used for the operation quantization:
    default_configuration_options = hw_model.QuantizationConfigOptions([eight_bits])

    # Create a HardwareModel and set its default quantization config.
    # This default configuration will be used for all operations
    # unless specified otherwise (see OperatorsSet, for example):
    qnnpack_model = hw_model.HardwareModel(default_configuration_options, name='qnnpack')

    # To start defining the model's components (such as operator sets, and fusing patterns),
    # use 'with' the hardware model instance, and create them as below:
    with qnnpack_model:
        # Combine operations/modules into a single module.
        # Pytorch supports the next fusing patterns:
        # [Conv, Relu], [Conv, BatchNorm], [Conv, BatchNorm, Relu], [Linear, Relu]
        # Source: # https://pytorch.org/docs/stable/quantization.html#model-preparation-for-quantization-eager-mode
        conv = hw_model.OperatorsSet("Conv")
        batchnorm = hw_model.OperatorsSet("BatchNorm")
        relu = hw_model.OperatorsSet("Relu")
        linear = hw_model.OperatorsSet("Linear")

        hw_model.Fusing([conv, batchnorm, relu])
        hw_model.Fusing([conv, batchnorm])
        hw_model.Fusing([conv, relu])
        hw_model.Fusing([linear, relu])

    return qnnpack_model




# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Callable, List, Tuple

import tensorflow as tf
from tqdm import tqdm

from model_compression_toolkit import common
from model_compression_toolkit.common.gptq.gptq_config import GradientPTQConfig
from model_compression_toolkit.common import Graph
from model_compression_toolkit.keras.back2framework.model_builder import model_builder
from model_compression_toolkit.common.model_builder_mode import ModelBuilderMode
from model_compression_toolkit.keras.gradient_ptq.graph_info import get_compare_points, \
    get_trainable_parameters, get_weights_for_loss
from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.keras.gradient_ptq.graph_update import update_graph_after_gptq
import numpy as np


def gptq_training_wrapper(tg: Graph,
                          graph_bias: Graph,
                          representative_data_gen: Callable,
                          gptq_config: GradientPTQConfig,
                          fw_info: FrameworkInfo) -> Graph:
    """
    Build two models from a graph: A teacher network (float model) and a student network (quantized model).
    Use the dataset generator to pass images through the teacher and student networks to get intermediate
    layers outputs. Use the outputs to compute the observed loss and to backpropagate the error
    in the student network, to minimize it in the next similar steps.
    All parameters (such as number of iterations, optimizer, etc.) are in the passed GradientPTQConfig.

    Args:
        tg: Graph to build a float networks from.
        graph_bias: Graph to build a quantized networks from (includes bias correction).
        representative_data_gen: Dataset generator to get images.
        gptq_config: GradientPTQConfig with parameters about the tuning process.
        fw_info: Framework information needed for keras kernel ops list.

    Returns:
        Graph of quantized model after GPTQ training.
    """

    input_scale = 1
    #########################################
    # Build two models and compare points
    #########################################
    compare_points, _, compare_points_mean, compare_points_std = get_compare_points(tg)  # get compare points

    float_model, float_user_info = model_builder(tg,
                                                 mode=ModelBuilderMode.FLOAT,
                                                 append2output=compare_points,
                                                 fw_info=fw_info)
    fxp_model, gptq_user_info = model_builder(graph_bias,
                                              mode=ModelBuilderMode.GPTQ,
                                              append2output=compare_points,
                                              fw_info=fw_info,
                                              gptq_config=gptq_config)

    trainable_weights = get_trainable_parameters(fxp_model,
                                                 fw_info,
                                                 add_bias=gptq_config.train_bias)
    flp_weights_list, fxp_weights_list = get_weights_for_loss(fxp_model)

    if not (len(compare_points) == len(trainable_weights) == len(flp_weights_list) == len(fxp_weights_list)):
        raise Exception("GPTQ: Mismatch between number of compare points, number of layers with trainable weights " +
                        "and number of float and quantized weights for loss")

    flattened_trainable_weights = [w for layer_weights in trainable_weights for w in layer_weights]

    if float_user_info.input_scale != gptq_user_info.input_scale:
        common.Logger.error("Input scale mismatch between float and GPTQ networks")  # pragma: no cover
    else:
        input_scale = gptq_user_info.input_scale
    #########################################
    # Optimization Loop
    #########################################

    def update_step(input_data: List[np.ndarray]) -> Tuple[tf.Tensor, List[tf.Tensor]]:
        """
        Get outputs from both teacher and student networks. Compute the observed error,
        and use it to compute the gradients and applying them to the student weights.
        Args:
            input_data: A list of Input tensors to pass through the networks.

        Returns:
            Loss and gradients.
        """
        y_float = float_model(input_data)  # running float model
        with tf.GradientTape(persistent=True) as tape:
            y_fxp = fxp_model(input_data)  # running fxp model
            loss_value = gptq_config.loss(y_fxp, y_float, fxp_weights_list, flp_weights_list,
                                          compare_points_mean, compare_points_std)

        # Use the gradient tape to automatically retrieve
        # the gradients of the trainable variables with respect to the loss.
        grads = tape.gradient(loss_value, flattened_trainable_weights)

        # Run one step of gradient descent by updating
        # the value of the variables to minimize the loss.
        gptq_config.optimizer.apply_gradients(zip(grads, flattened_trainable_weights))

        return loss_value, grads

    loss_list = []
    for _ in tqdm(range(gptq_config.n_iter)):
        data = representative_data_gen()
        loss_value_step, grads = update_step([d * input_scale for d in data])
        if gptq_config.log_function is not None:
            gptq_config.log_function(loss_value_step, grads, flattened_trainable_weights, compare_points)
        loss_list.append(loss_value_step.numpy())
        common.Logger.debug(f'last loss value: {loss_list[-1]}')

    #########################################
    # Update Graph after GPTQ
    #########################################
    tg_gptq = update_graph_after_gptq(fxp_model,
                                      graph_bias,
                                      add_bias=gptq_config.train_bias)

    tg_gptq.user_info.gptq_info_dict['loss'] = loss_list
    return tg_gptq



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import importlib

from model_compression_toolkit.common.constants import TENSORFLOW, PYTORCH
from model_compression_toolkit.common.hardware_representation import FrameworkHardwareModel


#############################
# Build Tensorflow models:
#############################

found_tf = importlib.util.find_spec("tensorflow") is not None and importlib.util.find_spec("tensorflow_model_optimization") is not None
tf_models_dict = {}

if found_tf:
    from model_compression_toolkit.hardware_models.keras_hardware_model.keras_default import get_default_hwm_keras
    from model_compression_toolkit.hardware_models.keras_hardware_model.keras_tflite import get_keras_hardware_model_tflite
    from model_compression_toolkit.hardware_models.keras_hardware_model.keras_qnnpack import get_qnnpack_tensorflow
    from model_compression_toolkit.keras.constants import DEFAULT_HWM, TFLITE_HWM, QNNPACK_HWM

    tf_models_dict = {DEFAULT_HWM: get_default_hwm_keras(),
                      TFLITE_HWM: get_keras_hardware_model_tflite(),
                      QNNPACK_HWM: get_qnnpack_tensorflow()}


#############################
# Build Pytorch models:
#############################
found_torch = importlib.util.find_spec("torch") is not None
torch_models_dict = {}

if found_torch:
    from model_compression_toolkit.hardware_models.pytorch_hardware_model.pytorch_default import get_default_hwm_pytorch
    from model_compression_toolkit.hardware_models.pytorch_hardware_model.pytorch_qnnpack import get_qnnpack_pytorch
    from model_compression_toolkit.hardware_models.pytorch_hardware_model.pytorch_tflite import get_pytorch_tflite_model
    from model_compression_toolkit.pytorch.constants import DEFAULT_HWM, TFLITE_HWM, QNNPACK_HWM

    torch_models_dict = {DEFAULT_HWM: get_default_hwm_pytorch(),
                         TFLITE_HWM: get_pytorch_tflite_model(),
                         QNNPACK_HWM: get_qnnpack_pytorch()}


fw_hw_models_dict = {TENSORFLOW: tf_models_dict,
                     PYTORCH: torch_models_dict}


def get_model(fw_name: str,
              hw_name: str) -> FrameworkHardwareModel:
    """
    Get a FrameworkHardwareModel by the hardware model name and the framework name.
    For now, it supports frameworks 'tensorflow' and 'pytorch'. For both of them
    the hardware model can be 'default','tflite', or 'qnnpack'.

    Args:
        fw_name: Framework name of the FrameworkHardwareModel.
        hw_name: Hardware model name the model will use for inference.

    Returns:
        A FrameworkHardwareModel object that models the hardware and attaches
        a framework information to it.
    """
    assert fw_name in fw_hw_models_dict, f'Framework {fw_name} is not supported'
    supported_models_by_fw = fw_hw_models_dict.get(fw_name)
    assert hw_name in supported_models_by_fw, f'Hardware model named {hw_name} is not' \
                                              f' supported for framework {fw_name}'
    return fw_hw_models_dict.get(fw_name).get(hw_name)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import copy

from typing import Any, List

from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig
from model_compression_toolkit.common import Graph, BaseNode
from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.common.logger import Logger
from model_compression_toolkit.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfig


def set_bit_widths(quant_config: QuantizationConfig,
                   graph_to_set_bit_widths: Graph,
                   fw_info: FrameworkInfo = None,
                   bit_widths_config: List[int] = None) -> Graph:
    """
    Set bit widths configuration to nodes in a graph. For each node, use the desired index
    in bit_widths_config to finalize the node weights and activation quantization configuration.

    Args:
        quant_config: MixedPrecisionQuantizationConfig the graph was computed according to.
        graph_to_set_bit_widths: A prepared for quantization graph to set its bit widths.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel
        channels indices, groups of layers by how they should be quantized, etc.)
        bit_widths_config: MP configuration (a list of indices: one for each node's candidate
        quantization configuration).

    """
    graph = copy.deepcopy(graph_to_set_bit_widths)

    if isinstance(quant_config, MixedPrecisionQuantizationConfig):
        assert all([len(n.candidates_quantization_cfg) > 0 for n in graph.get_configurable_sorted_nodes()]), \
            "All configurable nodes in graph should have at least one candidate configuration in mixed precision mode"

        Logger.info(f'Set bit widths from configuration: {bit_widths_config}')
        # Get a list of nodes' names we need to finalize (that they have at least one weight qc candidate).
        sorted_nodes_names = graph.get_configurable_sorted_nodes_names()
        for node in graph.nodes:  # set a specific node qc for each node final weights qc
            # If it's reused, take the configuration that the base node has
            node_name = node.name if not node.reuse else '_'.join(node.name.split('_')[:-2])
            if node_name in sorted_nodes_names:  # only configurable nodes are in this list
                node_index_in_graph = sorted_nodes_names.index(node_name)
                _set_node_final_qc(bit_widths_config,
                                   fw_info,
                                   node,
                                   node_index_in_graph)
            else:
                # TODO: refactor after adding activation mixed precision
                if node.is_activation_quantization_enabled():
                    assert node.is_all_activation_candidates_equal, "Activation Mixed Precision is not supported"
                    node.final_activation_quantization_cfg = node.candidates_quantization_cfg[
                        0].activation_quantization_cfg

    # When working in non-mixed-precision mode, there's only one bitwidth, and we simply set the
    # only candidate of the node as its final weight and activation quantization configuration.
    else:
        for n in graph.nodes:
            assert len(n.candidates_quantization_cfg) == 1
            n.final_weights_quantization_cfg = n.candidates_quantization_cfg[0].weights_quantization_cfg
            n.final_activation_quantization_cfg = n.candidates_quantization_cfg[0].activation_quantization_cfg

    return graph


def _get_node_qc_by_bit_widths(node: BaseNode,
                               bit_width_cfg: List[int],
                               node_index_in_graph: int) -> Any:
    """
    Get the node's quantization configuration that
    matches to the bit width index as in the MP configuration bit_width_cfg.
    If it was not found, return None.

    Args:
        node: Node to get its quantization configuration candidate.
        bit_width_cfg: Configuration which determines the node's desired bit width.
        node_index_in_graph: Index of the node in the bit_width_cfg.

    Returns:
        Node quantization configuration if it was found, or None otherwise.
    """

    if node.is_weights_quantization_enabled() or node.is_activation_quantization_enabled():
        bit_index_in_cfg = bit_width_cfg[node_index_in_graph]
        qc = node.candidates_quantization_cfg[bit_index_in_cfg]
        return qc
    return None


def _set_node_final_qc(bit_width_cfg: List[int],
                       fw_info: FrameworkInfo,
                       node: BaseNode,
                       node_index_in_graph: int):
    """
    Get the node's quantization configuration that
    matches to the bit width index as in the MP configuration bit_width_cfg, and use it to finalize the node's
    weights and activation quantization config.
    If the node quantization config was not found, raise an exception.

    Args:
        bit_width_cfg: Configuration which determines the node's desired bit width.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
        groups of layers by how they should be quantized, etc.)
        node: Node to set its node quantization configuration.
        node_index_in_graph: Index of the node in the bit_width_cfg.

    """
    node_qc = _get_node_qc_by_bit_widths(node, bit_width_cfg, node_index_in_graph)
    if node_qc is None:
        Logger.critical(f'Node {node.name} quantization configuration from configuration file'
                        f' was not found in candidates configurations.')
    else:
        node.final_weights_quantization_cfg = node_qc.weights_quantization_cfg
        node.final_activation_quantization_cfg = node_qc.activation_quantization_cfg



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import copy
import tensorflow as tf

# As from Tensorflow 2.6, keras is a separate package and some classes should be imported differently.
if tf.__version__ < "2.6":
    from tensorflow.python.keras.engine.base_layer import TensorFlowOpLayer
else:
    from keras.engine.base_layer import TensorFlowOpLayer

from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper

from model_compression_toolkit import common
from model_compression_toolkit.keras.constants import BIAS, USE_BIAS
from model_compression_toolkit.keras.quantizer.gradient_ptq import WeightQuantizeConfig
from tensorflow.keras.models import Model
from model_compression_toolkit.common.graph.base_graph import Graph


def update_graph_after_gptq(fxp_model: Model,
                            graph: Graph,
                            add_bias: bool = False) -> Graph:
    """
    After training using GPTQ, update nodes in a graph with thresholds and weights
    of their corresponding layers in a model (weights are updated for Conv2D and DepthwiseConv2D nodes only).

    Args:
        fxp_model: Model to gets updates from.
        graph: Graph to update its nodes.
        add_bias: Whether to update the nodes biases or not.

    Returns:
        An updated graph.
    """

    graph = copy.copy(graph)

    for layer in fxp_model.layers:
        if isinstance(layer, QuantizeWrapper) and isinstance(
                layer.quantize_config, WeightQuantizeConfig):
            node = graph.find_node_by_name(layer.layer.name)
            if len(node) == 0 and isinstance(layer.layer, TensorFlowOpLayer):
                node = graph.find_node_by_name('_'.join(layer.layer.name.split('_')[3:]))
            if len(node) != 1:
                common.Logger.error(f"Can't update GPTQ graph due to missing layer named: {layer.layer.name}")
            node = node[0]
            weights, weight_quant_config, activation_quant_config = \
                layer.quantize_config.update_layer_quantization_params(layer)
            for weight_attr, weight in weights.items():
                node.set_weights_by_keys(weight_attr, weight.numpy())
            for config_attr, config_value in weight_quant_config.items():
                node.final_weights_quantization_cfg.set_quant_config_attr(config_attr, config_value)
            for config_attr, config_value in activation_quant_config.items():
                node.final_activation_quantization_cfg.set_quant_config_attr(config_attr, config_value)
            if add_bias:
                use_bias = layer.layer.get_config().get(USE_BIAS)
                if use_bias is not None and use_bias:
                    new_bias = layer.layer.bias.numpy()
                    node.set_weights_by_keys(BIAS, new_bias)

    return graph



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import tensorflow as tf
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper
from typing import Tuple, List

from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.graph.base_node import BaseNode
from model_compression_toolkit.keras.constants import USE_BIAS
from model_compression_toolkit.keras.quantizer.gradient_ptq import WeightQuantizeConfig
from model_compression_toolkit.common.framework_info import FrameworkInfo
from tensorflow.keras.models import Model


def get_compare_points(input_graph: Graph) -> Tuple[List[BaseNode], List[str], List, List]:
    """
    Create a list of nodes with weights in a graph and a corresponding list
    of their names for tensors comparison purposes. Also outputs 2 list of activations
    prior information collected from batch normalization nodes (if exists)
    Args:
        input_graph: Graph to get its points to compare.

    Returns:
        A list of nodes in a graph
        A list of the their names.
        A list of nodes mean collected from BatchNorms in the graph
        A list of nodes std collected from BatchNorms in the graph
    """
    compare_points = []
    compare_points_mean = []
    compare_points_std = []
    compare_points_name = []
    for n in input_graph.get_topo_sorted_nodes():
        if len(n.weights) > 0:
            compare_points.append(n)
            compare_points_name.append(n.name)
            compare_points_std.append(n.prior_info.std_output)
            compare_points_mean.append(n.prior_info.mean_output)
    return compare_points, compare_points_name, compare_points_mean, compare_points_std


def get_trainable_parameters(fxp_model: Model,
                             fw_info: FrameworkInfo,
                             add_bias: bool = False) -> List[List[tf.Variable]]:
    """
    Get trainable parameters from all layers in a model

    Args:
        fxp_model: Model to get its trainable parameters.
        fw_info: Framework information needed for keras kernel ops list.
        add_bias: Whether to include biases of the model (if there are) or not.

    Returns:
        A list of trainable variables in a model. Each item is a list of a layers weights.
    """

    trainable_weights = []
    for layer in fxp_model.layers:
        if isinstance(layer, QuantizeWrapper) and isinstance(
                layer.quantize_config, WeightQuantizeConfig):
            # collect trainable weights per layer
            layer_trainable_weights = layer.quantize_config.get_trainable_quantizer_parameters()
            if add_bias:
                use_bias = isinstance(layer.layer, tuple(fw_info.kernel_ops)) and layer.layer.get_config().get(USE_BIAS)
                if use_bias is not None and use_bias:
                    layer_trainable_weights.append(layer.layer.bias)
            trainable_weights.append(layer_trainable_weights)

    return trainable_weights


def get_weights_for_loss(fxp_model: Model) -> Tuple[List[list], List[list]]:
    """
    Get all float and quantized kernels for the GPTQ loss

    Args:
        fxp_model: Model to get its float and quantized weights.

    Returns:
        A list of float kernels, each item is the float kernel of the layer
        A list of quantized kernels, each item is the quantized kernel of the layer
    """

    flp_weights_list = []
    fxp_weights_list = []
    for layer in fxp_model.layers:
        if isinstance(layer, QuantizeWrapper) and isinstance(
                layer.quantize_config, WeightQuantizeConfig):

            # collect pairs of float and quantized weights per layer
            _layer_flp_weights, _layer_fxp_weights = [], []
            for weight, quantizer, quantizer_vars in layer._weight_vars:
                _layer_flp_weights.append(weight)
                _layer_fxp_weights.append(quantizer(weight, training=False, weights=quantizer_vars))
            flp_weights_list.append(_layer_flp_weights)
            fxp_weights_list.append(_layer_fxp_weights)

    return flp_weights_list, fxp_weights_list



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf

# As from Tensorflow 2.6, keras is a separate package and some classes should be imported differently.
from model_compression_toolkit.common.hardware_representation import QuantizationMethod
from model_compression_toolkit.common.substitutions.shift_negative_activation import apply_shift_negative_correction

if tf.__version__ < "2.6":
    from tensorflow.python.keras.engine.base_layer import TensorFlowOpLayer
else:
    from keras.engine.base_layer import TensorFlowOpLayer

import numpy as np
from tensorflow.keras.layers import Activation, Conv2D, Dense, DepthwiseConv2D, ZeroPadding2D, Reshape, \
    GlobalAveragePooling2D, Dropout, ReLU, PReLU, ELU
from typing import Tuple, Any

from model_compression_toolkit import common, QuantizationConfig, FrameworkInfo
from model_compression_toolkit.common import BaseNode, Graph
from model_compression_toolkit.common.constants import FLOAT_32, DATA_TYPE
from model_compression_toolkit.keras.constants import NEGATIVE_SLOPE, PADDING, PAD_SAME, PAD_VALID, BIAS, USE_BIAS
from model_compression_toolkit.common.graph.graph_matchers import EdgeMatcher
from model_compression_toolkit.common.graph.graph_matchers import NodeOperationMatcher, \
    NodeFrameworkAttrMatcher

from model_compression_toolkit.keras.constants import KERNEL_SIZE, STRIDES, ACTIVATION, TRAINABLE, LAYER_NAME, SWISH, \
    SELU, GELU

# Tensorflow Op layer attributes:
NODE_DEF = 'node_def'
CONSTANTS = 'constants'

# NodeDef keys constants:
NODE_NAME = 'name'
NODE_DICT_TYPES = 'attr'
NODE_INPUT = 'input'
INPUT_VARIABLE_SUFFIX = '/y'

# NodeDef operators:
NODE_OP = 'op'
NODE_ADD_OPERATOR = 'Add'
NODE_PAD_OPERATOR = 'PadV2'

# NodeDef padding input variables names:
NODE_PAD_SIZE_NAME = 'padding'
NODE_PAD_VALUE_NAME = 'constant_values'
NODE_CONSTANTS_TYPE = 'type'
NODE_CONSTANTS_DT_FLOAT = 'DT_FLOAT'
NODE_CONSTANTS_DT_INT32 = 'DT_INT32'

SHIFT_NEGATIVE_NON_LINEAR_NUM_BITS = 16

"""
This substitution aims to solve an issue of activation with negative outputs where
the portion of the negative range is relatively small. In a symmetric quantization this causes 
of bit loosing as the entire negative quantization range does not contain
any values. To solve it, we shift the output of the activation by the minimal output value (quantized) such
that all values after the shifting are positive. To correct the impact of such shifting, a correction
to the next linear node is computed and added to its bias term.
If the linear node pads the input tensor with zeros, we modify the padded value as well.  
"""


def shift_negative_activation_node_matchers():
    # Match activation nodes with negative outputs.
    snc_node = NodeOperationMatcher(tf.nn.silu) | \
               NodeOperationMatcher(tf.nn.swish) | \
               NodeOperationMatcher(tf.nn.leaky_relu) | \
               NodeOperationMatcher(tf.nn.selu) | \
               NodeOperationMatcher(tf.nn.gelu) | \
               NodeOperationMatcher(tf.nn.elu) | \
               (NodeOperationMatcher(Activation) & (NodeFrameworkAttrMatcher(ACTIVATION, SWISH) |
                                                    NodeFrameworkAttrMatcher(ACTIVATION, GELU) |
                                                    NodeFrameworkAttrMatcher(ACTIVATION, SELU))) | \
               NodeOperationMatcher(PReLU) | \
               NodeOperationMatcher(ELU) | \
               (NodeOperationMatcher(ReLU) & NodeFrameworkAttrMatcher(NEGATIVE_SLOPE, 0.0).logic_not())  # Leaky ReLU

    # Match linear layers where we can add a correction.
    linear_node = NodeOperationMatcher(Conv2D) | \
                  NodeOperationMatcher(Dense) | \
                  NodeOperationMatcher(DepthwiseConv2D)

    # Match nodes that can be in between the non-linear node to the linear node,
    # and still the substitution can be applied correctly.
    bypass_node = NodeOperationMatcher(Reshape) | \
                  NodeOperationMatcher(GlobalAveragePooling2D) | \
                  NodeOperationMatcher(Dropout)

    # Match a pad node that can be in between the non-linear node to the linear node.
    pad_node = NodeOperationMatcher(ZeroPadding2D)

    return snc_node, linear_node, bypass_node, pad_node



def create_add_node(add_value: float,
                    prev_node_name: str,
                    input_shape: tuple) -> BaseNode:
    """
    Create a new Add node, with a constant to add.
    The name of the node is determined by its previous node's name.

    Args:
        add_value: Constant to add to to the node's input tensor.
        prev_node_name: The name of the node before the Add node
        input_shape: Shape of the Add node's input tensor.

    Returns:
        New Add node.
    """

    add_node_name = prev_node_name + '_post_add'

    fw_attr = {
        LAYER_NAME: add_node_name,
        TRAINABLE: False,
        DATA_TYPE: FLOAT_32,

        NODE_DEF: {
            NODE_NAME: add_node_name,
            NODE_DICT_TYPES: {'T': {NODE_CONSTANTS_TYPE: NODE_CONSTANTS_DT_FLOAT}},
            NODE_OP: NODE_ADD_OPERATOR,
            NODE_INPUT: [prev_node_name, add_node_name + INPUT_VARIABLE_SUFFIX],
        },
        CONSTANTS: {1: np.array(add_value, dtype=np.float32).reshape([1]*len(input_shape))}}

    add_node = common.graph.BaseNode(add_node_name,
                                     fw_attr,
                                     input_shape,
                                     input_shape,
                                     weights={},
                                     quantization_attr={},
                                     layer_class=TensorFlowOpLayer)
    return add_node


def create_pad_node(next_node_name: str,
                    prev_node_name: str,
                    value_to_pad: float,
                    input_shape: tuple,
                    pad_top: int,
                    pad_btm: int,
                    pad_left: int,
                    pad_right: int) -> BaseNode:
    """
    Create a pad node with a constant value to pad its input tensor.

    Args:
        next_node_name: Name of the node next to the pad node.
        prev_node_name: Name of the node previous to the pad node.
        value_to_pad: Constant to use for padding the input of the node.
        input_shape: Shape of input tensor.
        pad_top: Number of elements to pad above the tensor.
        pad_btm: Number of elements to pad below the tensor.
        pad_left: Number of elements to pad left to the tensor.
        pad_right: Number of elements to pad right to the tensor.

    Returns:
        A pad node which pads its input with a constant value.
    """

    pad_node_name = next_node_name + '_pre_pad'

    fw_attr = {LAYER_NAME: pad_node_name,
               TRAINABLE: False,
               DATA_TYPE: FLOAT_32,

               NODE_DEF: {
                   NODE_NAME: pad_node_name,
                   NODE_OP: NODE_PAD_OPERATOR,
                   NODE_INPUT: [prev_node_name,
                                pad_node_name + f'/{NODE_PAD_SIZE_NAME}',  # name of padding size variable
                                pad_node_name + f'/{NODE_PAD_VALUE_NAME}'],
                   NODE_DICT_TYPES: {'Tpaddings': {NODE_CONSTANTS_TYPE: NODE_CONSTANTS_DT_INT32},
                                     'T': {NODE_CONSTANTS_TYPE: NODE_CONSTANTS_DT_FLOAT}}},

               CONSTANTS: {1: np.array([[0, 0],
                                        [pad_top, pad_btm],
                                        [pad_left, pad_right],
                                        [0, 0]], dtype=np.int32),  # padding size
                           2: value_to_pad}}  # padded value

    padded_shape = list(input_shape)
    padded_shape[1] += pad_top + pad_btm
    padded_shape[2] += pad_left + pad_right
    pad_node = common.graph.BaseNode(pad_node_name,
                                     fw_attr,
                                     input_shape,
                                     tuple(padded_shape),
                                     weights={},
                                     quantization_attr={},
                                     layer_class=TensorFlowOpLayer)
    return pad_node


def compute_op2d_padding(op2d_node: BaseNode) -> Tuple[int, int, int, int]:
    """
    Compute the padding around an input tensor of a linear node.
    This is needed to replace tensorflow 'same' padding with actual number of elements to pad.

    Args:
        op2d_node: Node to compute the number of elements it adds when padding.

    Returns:
        Tuple of four numbers: number of elements to pad.
    """

    sh = op2d_node.framework_attr.get(STRIDES)[0]
    sw = op2d_node.framework_attr.get(STRIDES)[1]
    kh = op2d_node.framework_attr.get(KERNEL_SIZE)[0]
    kw = op2d_node.framework_attr.get(KERNEL_SIZE)[1]

    pad_along_h = sh * op2d_node.output_shape[1] - op2d_node.input_shape[1] + kh - sh
    pad_along_w = sw * op2d_node.output_shape[2] - op2d_node.input_shape[2] + kw - sw

    pad_top = pad_along_h // 2
    pad_btm = pad_along_h - (pad_along_h // 2)

    pad_left = pad_along_w // 2
    pad_right = pad_along_w - (pad_along_w // 2)

    return pad_top, pad_btm, pad_left, pad_right


def get_padding_values(op2d_node) -> Tuple[Any, Any]:
    """

    Args:
        op2d_node: convolution type node from which to extract the padding values.

    Returns:
        A tuple of containing the padding attribute and padding values.
    """
    padding, padding_values = None, None
    if op2d_node.framework_attr.get(PADDING) == PAD_SAME and not (
            op2d_node.framework_attr.get(KERNEL_SIZE)[0] == 1 and op2d_node.framework_attr.get(KERNEL_SIZE)[1] == 1):
        padding = compute_op2d_padding(op2d_node)
        padding_values = padding[0], padding[1], padding[2], padding[3]
        op2d_node.framework_attr[PADDING] = PAD_VALID
    return padding, padding_values


def is_padding_node_and_node_has_padding(pad_node_to_consider: BaseNode,
                                         next_node: BaseNode) -> bool:
    """

    Args:
        pad_node_to_consider: Pad node between the non-linear and linear nodes to consider when
        correcting the expected shift.
        next_node: The next node after the node in check for correction.

    Returns:
        Whether a padding node exists and the next node is a linear node with padding.
    """
    return pad_node_to_consider is not None and next_node.framework_attr.get(PADDING) == PAD_SAME


def keras_apply_shift_negative_correction(graph: Graph,
                                          quant_config: QuantizationConfig,
                                          fw_info: FrameworkInfo) -> Graph:
    """
    Apply shift negative correction (SNC) on a graph built from a Keras model.

    Args:
        graph: Graph to apply SNC on.
        quant_config: Quantization configuration.
        fw_info: FrameworkInfo object with information about the specific framework's module.

    Returns:
        Graph after SNC.
    """
    snc_node, linear_node, bypass_node, pad_node = shift_negative_activation_node_matchers()

    return apply_shift_negative_correction(graph,
                                           quant_config,
                                           fw_info,
                                           snc_node,
                                           linear_node,
                                           bypass_node,
                                           pad_node,
                                           create_add_node,
                                           get_padding_values,
                                           create_pad_node,
                                           is_padding_node_and_node_has_padding,
                                           PADDING,
                                           BIAS,
                                           USE_BIAS
                                           )


# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

def get_current_model():
    """

    Returns: The current HardwaeModel that is being used and accessed.

    """
    return _current_hardware_model.get()


class CurrentHardwareModel:
    """
    Wrapper of the current HardwareModel object that is being accessed and defined.
    """

    def __init__(self):
        super(CurrentHardwareModel, self).__init__()
        self.hwm = None

    def get(self):
        """

        Returns: The current HardwareModel that is being defined.

        """
        if self.hwm is None:
            raise Exception('Hardware model is not initialized.')
        return self.hwm

    def reset(self):
        """

        Reset the current HardwareModel so a new HardwareModel can be wrapped and
        used as the current HardwareModel object.

        """
        self.hwm = None

    def set(self, hwm):
        """
        Set and wrap a HardwareModel as the current HardwareModel.

        Args:
            hwm: HardwareModel to set as the current HardwareModel to access and use.

        """
        self.hwm = hwm


# Use a single instance for the current model.
_current_hardware_model = CurrentHardwareModel()



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Callable
from typing import Dict, List

from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.mixed_precision.kpi import KPI
from model_compression_toolkit.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfig
from model_compression_toolkit.common.framework_info import FrameworkInfo


class MixedPrecisionSearchManager(object):
    """
    Class to wrap and manage the search process of a mixed-precision configuration.
    """

    def __init__(self,
                 graph: Graph,
                 qc: MixedPrecisionQuantizationConfig,
                 fw_info: FrameworkInfo,
                 get_sensitivity_evaluation: Callable):
        """

        Args:
            graph: Graph to search for its MP configuration.
            qc: Quantization configuration for how the graph should be quantized.
            fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize).
            get_sensitivity_evaluation: Framework specific function to retrieve a metric computation function.
        """

        self.graph = graph
        self.qc = qc
        self.fw_info = fw_info
        self.get_sensitivity_evaluation = get_sensitivity_evaluation
        self.metrics_weights = self.qc.distance_weighting_method
        self.layer_to_bitwidth_mapping = self.get_search_space()
        self.compute_metric_fn = self.get_sensitivity_metric()

    def get_search_space(self) -> Dict[int, List[int]]:
        """
        The search space is a mapping from a node's index to a list of integers (possible bitwidths candidates indeces
        for the node).

        Returns:
            The entire search space of the graph.
        """

        indices_mapping = {}
        nodes_to_configure = self.graph.get_configurable_sorted_nodes()
        for idx, n in enumerate(nodes_to_configure):
            # For each node, get all possible bitwidth indices for it
            # (which is a list from 0 to the length of the candidates qc list of the node).
            indices_mapping[idx] = list(range(len(n.candidates_quantization_cfg)))  # all search_methods space
        return indices_mapping


    def get_sensitivity_metric(self) -> Callable:
        """

        Returns: Return a function (from the framework implementation) to compute a metric that
        indicates the similarity of the mixed-precision model (to the float model) for a given
        mixed-precision configuration.

        """
        # Get from the framework an evaluation function on how a MP configuration,
        # affects the expected loss.
        compute_metric_fn = self.get_sensitivity_evaluation(self.graph,
                                                            self.qc,
                                                            self.metrics_weights)
        return compute_metric_fn

    def get_kpi_metric(self) -> Callable:
        """

        Returns: A function to compute the KPI of a graph for a mixed-precision bitwidth
        configuration.

        """

        def _compute_kpi(mp_model_config: List[int]) -> KPI:
            """
            Compute and return the KPI of a graph for a given mixed-precision bitwidth
            configuration.

            Args:
                mp_model_config: Mixed-precision bitwidth configuration (list of integers).

            Returns:
                KPI of a model when using the passed mixed-precision configuration.

            """
            weights_memory = 0

            # Go over all nodes that shold be taken into consideration when computing the KPI.
            mp_nodes = self.graph.get_configurable_sorted_nodes_names()
            for n in self.graph.nodes:
                if n.name in mp_nodes:
                    node_idx = mp_nodes.index(n.name)
                    # TODO: modify to account for activations size when implementing activations mixed precision
                    node_nbits = n.candidates_quantization_cfg[mp_model_config[node_idx]].weights_quantization_cfg.weights_n_bits
                elif n.is_weights_quantization_enabled():
                    # The only valid way to get here is if the node is reused (which means that we're not looking
                    # for its configuration), and we ignore it when computing the KPI (as the base node will acount
                    # for it).
                    assert n.reuse, "If node has candidates it should be part of the configurable nodes," \
                                    " unless it's a reused node"
                    node_nbits = 0  # Ignore reused nodes is the KPI computation.
                else:  # No weights quantization
                    node_nbits = 0
                node_num_params = 0

                # Consider only the weights that should be quantized.
                for attr in self.fw_info.get_kernel_op_attributes(n.type):
                    if attr is not None:
                        node_num_params += n.get_weights_by_keys(attr).flatten().shape[0]

                node_memory_in_bytes = node_num_params * node_nbits / 8.0
                weights_memory += node_memory_in_bytes

            return KPI(weights_memory=weights_memory)

        return _compute_kpi



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit.common.hardware_representation.hardware2framework import \
    FrameworkHardwareModel, LayerFilterParams
from model_compression_toolkit.common.hardware_representation.hardware2framework import \
    OperationsSetToLayers
from model_compression_toolkit.hardware_models.qnnpack import get_qnnpack_model

import tensorflow as tf
if tf.__version__ < "2.6":
    from tensorflow.keras.layers import Conv2D, DepthwiseConv2D, Conv2DTranspose, Dense, BatchNormalization, ReLU, Activation
else:
    from keras.layers import Conv2D, DepthwiseConv2D, Conv2DTranspose, Dense, BatchNormalization, ReLU, Activation


def get_qnnpack_tensorflow():
    qnnpackhm = get_qnnpack_model()
    qnnpack_tf = FrameworkHardwareModel(qnnpackhm,
                                        name='qnnpack_tensorflow')

    with qnnpack_tf:
        OperationsSetToLayers("Conv", [Conv2D,
                                       DepthwiseConv2D,
                                       Conv2DTranspose,
                                       tf.nn.conv2d,
                                       tf.nn.depthwise_conv2d,
                                       tf.nn.conv2d_transpose])

        OperationsSetToLayers("Linear", [Dense])

        OperationsSetToLayers("BatchNorm", [BatchNormalization,
                                            tf.nn.batch_normalization])

        OperationsSetToLayers("Relu", [tf.nn.relu,
                                       tf.nn.relu6,
                                       LayerFilterParams(ReLU, negative_slope=0.0),
                                       LayerFilterParams(Activation, activation="relu")])

    return qnnpack_tf





# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import copy
from typing import List, Dict

from model_compression_toolkit.common import Logger, BaseNode
from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfig
from model_compression_toolkit.common.quantization.candidate_node_quantization_config import \
    CandidateNodeQuantizationConfig
from model_compression_toolkit.common.quantization.node_quantization_config import NodeActivationQuantizationConfig
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig
from model_compression_toolkit.common.quantization.quantization_params_fn_selection import \
    get_activation_quantization_params_fn, get_weights_quantization_params_fn
from model_compression_toolkit.common.hardware_representation.hardware2framework import FrameworkHardwareModel
from model_compression_toolkit.common.hardware_representation.op_quantization_config import OpQuantizationConfig, \
    QuantizationConfigOptions


def set_quantization_configuration_to_graph(graph: Graph,
                                            quant_config: QuantizationConfig) -> Graph:
    """
    Add quantization configuration for each graph node.

    Args:
        graph: Graph for which to add quantization info to each node.
        quant_config: Quantization configuration containing parameters for how the graph should be quantized.

    Returns:
        The graph with quantization configurations attached to each node in it.
    """

    graph_with_qcs = copy.deepcopy(graph)
    for n in graph_with_qcs.nodes:
        set_quantization_configs_to_node(node=n,
                                         quant_config=quant_config,
                                         fw_info=graph.fw_info,
                                         fw_hw_model=graph.fw_hw_model)
    return graph_with_qcs


def set_quantization_configs_to_node(node: BaseNode,
                                     quant_config: QuantizationConfig,
                                     fw_info: FrameworkInfo,
                                     fw_hw_model: FrameworkHardwareModel):
    """
    Create and set quantization configurations to a node (for both weights and activation).

    Args:
        node: Node to set its quantization configurations.
        quant_config: Quantization configuration to generate the node's configurations from.
        fw_info: Information needed for quantization about the specific framework.
        fw_hw_model: FrameworkHardwareModel to get default OpQuantizationConfig.

    """
    node_qc_options = fw_hw_model.get_qco_by_node(node)

    # Create QC candidates for weights and activation combined
    weight_channel_axis = fw_info.kernel_channels_mapping.get(node.type)[0]
    node.candidates_quantization_cfg = _create_node_candidates_qc(quant_config,
                                                                  fw_info,
                                                                  weight_channel_axis,
                                                                  node_qc_options)

    enable_activation_quantization = quant_config.enable_activation_quantization \
                                     and (fw_info.in_activation_ops(node) or fw_info.in_kernel_ops(node)) \
                                     and node.get_has_activation()
    enable_weights_quantization = quant_config.enable_weights_quantization \
                                  and fw_info.in_kernel_ops(node) \
                                  and node.has_weights_to_quantize(fw_info)
    for candidate_qc in node.candidates_quantization_cfg:
        candidate_qc.weights_quantization_cfg.enable_weights_quantization = enable_weights_quantization
        candidate_qc.activation_quantization_cfg.enable_activation_quantization = enable_activation_quantization


def create_node_activation_qc(qc: QuantizationConfig,
                              fw_info: FrameworkInfo,
                              op_cfg: OpQuantizationConfig) -> NodeActivationQuantizationConfig:
    """
    Create an activation quantization configuration from a QuantizationConfig object.

    Args:
        qc: QuantizationConfig to create the node's config from.
        fw_info: Information about the specific framework the node was created from (e.g., whether or not its
        weights/activations should be quantized)
        op_cfg: OpQuantizationConfig with quantizers types to set in node quantization configuration.

    Returns:
        Activation quantization configuration of a node.
    """

    activation_quantization_fn = fw_info.activation_quantizer_mapping.get(op_cfg.activation_quantization_method)
    if activation_quantization_fn is None:
        Logger.critical('Unknown quantization method for activations')

    activation_quantization_params_fn = get_activation_quantization_params_fn(op_cfg.activation_quantization_method,
                                                                              qc.activation_error_method)

    return NodeActivationQuantizationConfig(qc,
                                            op_cfg,
                                            activation_quantization_fn,
                                            activation_quantization_params_fn)


def create_node_qc_candidate(qc: QuantizationConfig,
                             fw_info: FrameworkInfo,
                             weight_channel_axis: int,
                             op_cfg: OpQuantizationConfig) -> CandidateNodeQuantizationConfig:
    """
    Create quantization configuration candidate from a QuantizationConfig object.
    Creates both weights and activation quantization configurations
    and initialize a candidate object that encapsulates both.

    Args:
        qc: QuantizationConfig to create the node's config from.
        fw_info: Information about the specific framework the node was created from (e.g., whether its
            weights/activations should be quantized)
        weight_channel_axis: Output channel index of the node's kernel.
        op_cfg: OpQuantizationConfig of the node with quantizers types to use when creating node quantization configuration.

    Returns: a CandidateNodeQuantizationConfig object with both weights and activation quantization config objects.

    """

    # get attributes for weights quantization
    weights_quantization_fn = fw_info.weights_quantizer_mapping.get(op_cfg.weights_quantization_method)

    if weights_quantization_fn is None:
        Logger.critical('Unknown quantization method for weights')

    weights_quantization_params_fn = get_weights_quantization_params_fn(op_cfg.weights_quantization_method,
                                                                        qc.weights_error_method)

    # get attributes for activation quantization
    activation_quantization_fn = fw_info.activation_quantizer_mapping.get(op_cfg.activation_quantization_method)
    if activation_quantization_fn is None:
        Logger.critical('Unknown quantization method for activations')

    activation_quantization_params_fn = get_activation_quantization_params_fn(op_cfg.activation_quantization_method,
                                                                              qc.activation_error_method)

    return CandidateNodeQuantizationConfig(qc,
                                           op_cfg,
                                           activation_quantization_fn,
                                           activation_quantization_params_fn,
                                           weights_quantization_fn,
                                           weights_quantization_params_fn,
                                           weight_channel_axis)


def _create_node_candidates_qc(qc: QuantizationConfig,
                               fw_info: FrameworkInfo,
                               weight_channel_axis: int,
                               node_qc_options: QuantizationConfigOptions) -> List[CandidateNodeQuantizationConfig]:
    """
    Create a list of candidates of weights and activation quantization configurations for a node.

    Args:
        qc: Quantization configuration the quantization process should follow.
        fw_info: Framework information (e.g., which layers should have their kernels' quantized).
        weight_channel_axis: Output channel index of the node's kernel.
        node_qc_options: QuantizationConfigOptions for the node with quantization candidates information.

    Returns:
        List of candidates of weights quantization configurations to set for a node.
    """

    candidates = []
    if isinstance(qc, MixedPrecisionQuantizationConfig):
        for op_cfg in node_qc_options.quantization_config_list:
            candidate_nbits_qc = copy.deepcopy(qc)
            candidates.append(create_node_qc_candidate(candidate_nbits_qc,
                                                       fw_info,
                                                       weight_channel_axis,
                                                       op_cfg))
        # sorting the candidates by weights number of bits first and then by activation number of bits
        # (in reversed order)
        candidates.sort(key=lambda c: (c.weights_quantization_cfg.weights_n_bits,
                                       c.activation_quantization_cfg.activation_n_bits), reverse=True)
    else:
        op_cfg = node_qc_options.quantization_config_list[0]
        if len(node_qc_options.quantization_config_list) > 1:
            op_cfg = node_qc_options.base_config
            Logger.warning(f"Given multiple candidates configurations for non mixed-precision quantization,"
                           f"using base_config as the node's configuration")
        candidates.append(create_node_qc_candidate(qc,
                                                   fw_info,
                                                   weight_channel_axis,
                                                   op_cfg))

    return candidates



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Any

import numpy as np


#########################
#  Helpful functions
#########################


def validate_before_compute_similarity(float_tensor: Any, fxp_tensor: Any):
    """
    Assert different conditions before comparing two tensors (such as dimensionality and type).
    Args:
        float_tensor: First tensor to validate.
        fxp_tensor: Second tensor to validate.

    """
    assert isinstance(float_tensor, np.ndarray)
    assert isinstance(fxp_tensor, np.ndarray)
    assert float_tensor.shape == fxp_tensor.shape


def tensor_norm(x: np.ndarray, p: float = 2.0) -> np.float:
    """
    Compute the Lp-norm of a tensor x.
    Args:
        x: Tensor to compute its norm
        p: P to use for the Lp norm computation

    Returns:
        Lp norm of x.
    """
    return np.power(np.power(np.abs(x.flatten()), p).sum(), 1.0/p)


#########################
# Similarity functions
#########################

def compute_mse(float_tensor: np.ndarray, fxp_tensor: np.ndarray, norm: bool = False, norm_eps: float = 1e-8) -> float:
    """
    Compute the mean square error between two numpy arrays.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.
        norm: whether to normalize the error function result.
        norm_eps: epsilon value for error normalization stability.

    Returns:
        The MSE distance between the two tensors.
    """
    validate_before_compute_similarity(float_tensor, fxp_tensor)
    error = np.power(float_tensor - fxp_tensor, 2.0).mean()
    if norm:
        error /= (np.power(float_tensor, 2.0).mean() + norm_eps)
    return error


def compute_nmse(float_tensor: np.ndarray, fxp_tensor: np.ndarray) -> float:
    """
    Compute the normalized mean square error between two numpy arrays.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.

    Returns:
        The NMSE distance between the two tensors.
    """
    validate_before_compute_similarity(float_tensor, fxp_tensor)
    normalized_float_tensor = float_tensor / tensor_norm(float_tensor)
    normalized_fxp_tensor = fxp_tensor / tensor_norm(fxp_tensor)
    return np.mean(np.power(normalized_float_tensor - normalized_fxp_tensor, 2.0))


def compute_nmae(float_tensor: np.ndarray, fxp_tensor: np.ndarray) -> float:
    """
    Compute the normalized mean average error between two numpy arrays.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.

    Returns:
        The NMAE distance between the two tensors.
    """
    validate_before_compute_similarity(float_tensor, fxp_tensor)
    normalized_float_tensor = float_tensor / tensor_norm(float_tensor, 1.0)
    normalized_fxp_tensor = fxp_tensor / tensor_norm(fxp_tensor, 1.0)
    return np.mean(np.abs(normalized_float_tensor - normalized_fxp_tensor))


def compute_mae(float_tensor: np.ndarray, fxp_tensor: np.ndarray, norm: bool = False, norm_eps: float = 1e-8) -> float:
    """
    Compute the mean average error function between two numpy arrays.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.
        norm: whether to normalize the error function result.
        norm_eps: epsilon value for error normalization stability.

    Returns:
        The mean average distance between the two tensors.
    """

    validate_before_compute_similarity(float_tensor, fxp_tensor)
    error = np.abs(float_tensor - fxp_tensor).mean()
    if norm:
        error /= (np.abs(float_tensor).mean() + norm_eps)
    return error


def compute_cs(float_tensor: np.ndarray, fxp_tensor: np.ndarray, eps: float = 1e-8) -> float:
    """
    Compute the similarity between two tensor using cosine similarity.
    The returned values is between 0 to 1: the smaller returned value,
    the greater similarity there is between the two tensors.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.
        eps: Small value to avoid zero division.

    Returns:
        The cosine similarity between two tensors.
    """

    validate_before_compute_similarity(float_tensor, fxp_tensor)
    if np.all(fxp_tensor == 0) and np.all(float_tensor == 0):
        return 1.0

    float_flat = float_tensor.flatten()
    fxp_flat = fxp_tensor.flatten()
    float_norm = tensor_norm(float_tensor)
    fxp_norm = tensor_norm(fxp_tensor)

    # -1 <= cs <= 1
    cs = np.sum(float_flat * fxp_flat) / ((float_norm * fxp_norm) + eps)

    # Return a non-negative float (smaller value -> more similarity)
    return (1.0 - cs) / 2.0


def compute_lp_norm(float_tensor: np.ndarray, fxp_tensor: np.ndarray, p: int, norm: bool = False,
                    norm_eps: float = 1e-8) -> float:
    """
    Compute the error function between two numpy arrays.
    The error is computed based on Lp-norm distance of the tensors.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.
        p: p-norm to use for the Lp-norm distance.
        norm: whether to normalize the error function result.
        norm_eps: epsilon value for error normalization stability.

    Returns:
        The Lp-norm distance between the two tensors.
    """
    validate_before_compute_similarity(float_tensor, fxp_tensor)
    error = np.power(np.abs(float_tensor - fxp_tensor), p).mean()
    if norm:
        error /= (np.power(np.abs(float_tensor), p).mean() + norm_eps)
    return error



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable, List

from model_compression_toolkit import common
from model_compression_toolkit.common import Logger
from model_compression_toolkit.common.constants import PYTORCH
from model_compression_toolkit.common.gptq.gptq_config import GradientPTQConfig
from model_compression_toolkit.common.hardware_representation import FrameworkHardwareModel
from model_compression_toolkit.common.mixed_precision.kpi import KPI
from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.common.network_editors.actions import EditRule
from model_compression_toolkit.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfig, DEFAULT_MIXEDPRECISION_CONFIG
from model_compression_toolkit.common.post_training_quantization import post_training_quantization
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig
from model_compression_toolkit.common.quantization.quantization_config import DEFAULTCONFIG

import importlib

if importlib.util.find_spec("torch") is not None:
    from model_compression_toolkit.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
    from model_compression_toolkit.pytorch.pytorch_implementation import PytorchImplementation
    from model_compression_toolkit.pytorch.constants import DEFAULT_HWM
    from torch.nn import Module

    from model_compression_toolkit import get_model
    PYTORCH_DEFAULT_MODEL = get_model(PYTORCH, DEFAULT_HWM)

    def pytorch_post_training_quantization(in_module: Module,
                                           representative_data_gen: Callable,
                                           n_iter: int = 500,
                                           quant_config: QuantizationConfig = DEFAULTCONFIG,
                                           fw_info: FrameworkInfo = DEFAULT_PYTORCH_INFO,
                                           network_editor: List[EditRule] = [],
                                           gptq_config: GradientPTQConfig = None,
                                           analyze_similarity: bool = False,
                                           fw_hw_model: FrameworkHardwareModel = PYTORCH_DEFAULT_MODEL):
        """
        Quantize a trained Pytorch module using post-training quantization. The module is quantized using a
        symmetric constraint quantization thresholds (power of two).
        The module is first optimized using several transformations (e.g. BatchNormalization folding to
        preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
        being collected for each layer's output (and input, depends on the quantization configuration).
        Thresholds are then being calculated using the collected statistics and the module is quantized
        (both coefficients and activations by default).
        If a gptq configuration is passed, the quantized weights are optimized using gradient based post
        training quantization by comparing points between the float and quantized modules, and minimizing the observed loss.

        Args:
            in_module (Module): Pytorch module to quantize.
            representative_data_gen (Callable): Dataset used for calibration.
            n_iter (int): Number of calibration iterations to run.
            quant_config (QuantizationConfig): QuantizationConfig containing parameters of how the module should be quantized. `Default configuration. <https://github.com/sony/model_optimization/blob/21e21c95ca25a31874a5be7af9dd2dd5da8f3a10/model_compression_toolkit/common/quantization/quantization_config.py#L154>`_
            fw_info (FrameworkInfo): Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.). `Default Pytorch info <https://github.com/sony/model_optimization/blob/21e21c95ca25a31874a5be7af9dd2dd5da8f3a10/model_compression_toolkit/pytorch/default_framework_info.py#L113>`_
            network_editor (List[EditRule]): List of EditRules. Each EditRule consists of a node filter and an action to change quantization settings of the filtered nodes.
            gptq_config (GradientPTQConfig): Configuration for using gptq (e.g. optimizer).
            analyze_similarity (bool): Whether to plot similarity figures within TensorBoard (when logger is enabled) or not.
            fw_hw_model (FrameworkHardwareModel): FrameworkHardwareModel to optimize the Keras model according to.


        Returns:
            A quantized module and information the user may need to handle the quantized module.

        Examples:

            Import a Pytorch module:

            >>> import torchvision.models.mobilenet_v2 as models
            >>> module = models.mobilenet_v2()

            Create a random dataset generator:

            >>> import numpy as np
            >>> def repr_datagen(): return [np.random.random((1,224,224,3))]

            Import mct and pass the module with the representative dataset generator to get a quantized module:

            >>> import model_compression_toolkit as mct
            >>> quantized_module, quantization_info = mct.pytorch_post_training_quantization(module, repr_datagen)

        """

        return post_training_quantization(in_module,
                                          representative_data_gen,
                                          n_iter,
                                          quant_config,
                                          fw_info,
                                          PytorchImplementation(),
                                          fw_hw_model,
                                          network_editor,
                                          gptq_config,
                                          analyze_similarity)


    def pytorch_post_training_quantization_mixed_precision(in_model: Module,
                                                           representative_data_gen: Callable,
                                                           target_kpi: KPI,
                                                           n_iter: int = 500,
                                                           quant_config: MixedPrecisionQuantizationConfig = DEFAULT_MIXEDPRECISION_CONFIG,
                                                           fw_info: FrameworkInfo = DEFAULT_PYTORCH_INFO,
                                                           network_editor: List[EditRule] = [],
                                                           gptq_config: GradientPTQConfig = None,
                                                           analyze_similarity: bool = False,

                                                           fw_hw_model: FrameworkHardwareModel = PYTORCH_DEFAULT_MODEL):
        """
         Quantize a trained Pytorch model using post-training quantization. The model is quantized using a
         symmetric constraint quantization thresholds (power of two).
         The model is first optimized using several transformations (e.g. BatchNormalization folding to
         preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
         being collected for each layer's output (and input, depends on the quantization configuration).
         For each possible bit width (per layer) a threshold is then being calculated using the collected
         statistics. Then, using an ILP solver we find a mixed-precision configuration, and set a bit width
         for each layer. The model is then quantized (both coefficients and activations by default).
         In order to limit the maximal model's size, a target KPI need to be passed after weights_memory
         is set (in bytes).
         If a gptq configuration is passed, the quantized weights are optimized using gradient based post
         training quantization by comparing points between the float and quantized models, and minimizing the observed loss.
         Notice that this feature is experimental.
         **For now, mixed precision is supported for weights only.**

         Args:
             in_model (Model): Pytorch model to quantize.
             representative_data_gen (Callable): Dataset used for calibration.
             target_kpi (KPI): KPI object to limit the search of the mixed-precision configuration as desired.
             n_iter (int): Number of calibration iterations to run.
             quant_config (MixedPrecisionQuantizationConfig): QuantizationConfig containing parameters of how the model should be quantized.
             fw_info (FrameworkInfo): Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.). `Default Keras info <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/pytorch/default_framework_info.py#L100>`_
             network_editor (List[EditRule]): List of EditRules. Each EditRule consists of a node filter and an action to change quantization settings of the filtered nodes.
             gptq_config (GradientPTQConfig): Configuration for using GPTQ (e.g. optimizer).
             analyze_similarity (bool): Whether to plot similarity figures within TensorBoard (when logger is enabled) or not.
             fw_hw_model (FrameworkHardwareModel): FrameworkHardwareModel to optimize the Keras model according to.

         Returns:
             A quantized model and information the user may need to handle the quantized model.

         Examples:

             Import MCT:

             >>> import model_compression_toolkit as mct

             Import a Pytorch model:

             >>> import torchvision.models.mobilenet_v2 as models
             >>> module = models.mobilenet_v2()

             Create a random dataset generator:

             >>> import numpy as np
             >>> def repr_datagen(): return [np.random.random((1,224,224,3))]

             Create a mixed-precision configuration, to quantize a model with different bitwidths for different layers.
             The candidates bitwidth for quantization should be defined in the hardware model:

             >>> config = mct.MixedPrecisionQuantizationConfig()

             Create a KPI object to limit our returned model's size. Note that this value affects only coefficients that should be quantized (for example, the kernel of Conv2D in Keras will be affected by this value, while the bias will not):

             >>> kpi = mct.KPI(sum(p.numel() for p in module.parameters()) * 0.75)  # About 0.75 of the model size when quantized with 8 bits.

             Pass the model, the representative dataset generator, the configuration and the target KPI to get a quantized model:

             >>> quantized_model, quantization_info = mct.pytorch_post_training_quantization_mixed_precision(module, repr_datagen, n_iter=10, quant_config=config, target_kpi=kpi)

             For more configuration options, please take a look at our `API documentation <https://sony.github.io/model_optimization/api/api_docs/modules/mixed_precision_quantization_config.html>`_.

         """
        if not isinstance(quant_config, MixedPrecisionQuantizationConfig):
            common.Logger.error("Given quantization config to mixed-precision facade is not of type "
                                "MixedPrecisionQuantizationConfig. Please use pytorch_post_training_quantization API, "
                                "or pass a valid mixed precision configuration.")

        common.Logger.info("Using experimental mixed-precision quantization. "
                           "If you encounter an issue please file a bug.")

        return post_training_quantization(in_model,
                                          representative_data_gen,
                                          n_iter,
                                          quant_config,
                                          fw_info,
                                          PytorchImplementation(),
                                          fw_hw_model,
                                          network_editor,
                                          gptq_config,
                                          analyze_similarity,
                                          target_kpi)

else:
    # If torch is not installed,
    # we raise an exception when trying to use these functions.
    def pytorch_post_training_quantization(*args, **kwargs):
        Logger.critical('Installing Pytorch is mandatory '
                        'when using pytorch_post_training_quantization. '
                        'Could not find the torch package.')

    def pytorch_post_training_quantization_mixed_precision(*args, **kwargs):
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using pytorch_post_training_quantization_mixed_precision. '
                        'Could not find Tensorflow package.')



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np

from model_compression_toolkit.common.constants import RANGE_MIN, RANGE_MAX, THRESHOLD
from model_compression_toolkit.common.quantization.quantizers.quantizers_helpers import uniform_quantize_tensor, \
    quantize_tensor


def threshold_is_power_of_two(threshold: np.ndarray, per_channel: bool) -> bool:
    if per_channel:
        thresholds_per_channel = threshold.flatten()
        return (np.log2(thresholds_per_channel) == list(map(int, np.log2(thresholds_per_channel)))).all()

    return np.log2(threshold) == int(np.log2(threshold))


def power_of_two_quantizer(tensor_data: np.ndarray,
                           n_bits: int,
                           signed: bool,
                           quantization_params: dict,
                           per_channel: bool,
                           output_channels_axis: int) -> np.ndarray:
    """
    Quantize a tensor according to given: threshold, number of bits, and whether
    quantization range is sign or unsigned.

    Args:
        tensor_data: Tensor values to quantize.
        n_bits: Number of bits to quantize the tensor.
        signed: Whether the tensor contains negative values or not.
        quantization_params: Dictionary of specific parameters for this quantization function.
        per_channel: Whether to use separate quantization per output channel.
        output_channels_axis: Axis of the output channel.

    Returns:
        Quantized data.
    """
    threshold = quantization_params.get(THRESHOLD)
    if threshold is None:
        raise Exception(f"{THRESHOLD} parameter must be defined in 'quantization_params'")
    if not threshold_is_power_of_two(threshold, per_channel):
        raise Exception(f"Expects {THRESHOLD} parameter to be a power of two, but got {threshold}")

    return quantize_tensor(tensor_data,
                           threshold,
                           n_bits,
                           signed)


def symmetric_quantizer(tensor_data: np.ndarray,
                        n_bits: int,
                        signed: bool,
                        quantization_params: dict,
                        per_channel: bool,
                        output_channels_axis: int) -> np.ndarray:
    """
    Quantize a tensor in a symmetric range according to given: threshold, number of bits, and whether
    quantization range is symmetric and is sign or unsigned.

    Args:
        tensor_data: Tensor values to quantize.
        n_bits: Number of bits to quantize the tensor.
        signed: Whether the tensor contains negative values or not.
        quantization_params: Dictionary of specific parameters for this quantization function.
        per_channel: Whether to use separate quantization per output channel.
        output_channels_axis: Axis of the output channel.

    Returns:
        Quantized data.
    """
    threshold = quantization_params.get(THRESHOLD)
    if threshold is None:
        raise Exception(f"{THRESHOLD} parameter must be defined in 'quantization_params'")

    return quantize_tensor(tensor_data,
                           threshold,
                           n_bits,
                           signed)


def uniform_quantizer(tensor_data: np.ndarray,
                      n_bits: int,
                      signed: bool,
                      quantization_params: dict,
                      per_channel: bool,
                      output_channels_axis: int) -> np.ndarray:
    """
    Quantize a tensor uniformly in the given range.

    Args:
        tensor_data: Tensor values to quantize.
        n_bits: Number of bits to quantize the tensor.
        signed: Whether the tensor contains negative values or not (not used in uniform quantization).
        quantization_params: Dictionary of specific parameters for this quantization function.
        per_channel: Whether to use separate quantization per output channel.
        output_channels_axis: Axis of the output channel.

    Returns:
        Quantized data.
    """
    range_min = quantization_params.get(RANGE_MIN)
    range_max = quantization_params.get(RANGE_MAX)
    if range_min is None or range_max is None:
        raise Exception("'quantization range' parameters must be defined in 'quantization_params'")

    return uniform_quantize_tensor(tensor_data, range_min, range_max, n_bits)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
from torch.nn import Hardswish, Hardsigmoid, ReLU, Hardtanh, ReLU6, LeakyReLU, PReLU, SiLU, Softmax, \
    Sigmoid, Softplus, Softsign, Tanh
from torch.nn.functional import hardswish, hardsigmoid, relu, hardtanh, relu6, leaky_relu, prelu, silu, softmax, \
    softplus, softsign
from torch.nn import UpsamplingBilinear2d, AdaptiveAvgPool2d, AvgPool2d, MaxPool2d
from torch.nn.functional import upsample_bilinear, adaptive_avg_pool2d, avg_pool2d, max_pool2d
from torch.nn import Conv2d, ConvTranspose2d, Linear, BatchNorm2d
from torch.nn import Dropout, Flatten
from torch import add, multiply, mul, sub, flatten, reshape, split, unsqueeze, concat, cat,\
    mean, dropout, sigmoid, tanh
import operator

from model_compression_toolkit.common.defaultdict import DefaultDict
from model_compression_toolkit.common.framework_info import FrameworkInfo, ChannelAxis
from model_compression_toolkit.common.hardware_representation import QuantizationMethod
from model_compression_toolkit.common.quantization.quantizers.kmeans_quantizer import kmeans_quantizer
from model_compression_toolkit.common.quantization.quantizers.lut_kmeans_quantizer import lut_kmeans_quantizer
from model_compression_toolkit.common.quantization.quantizers.uniform_quantizers import power_of_two_quantizer, \
    symmetric_quantizer, uniform_quantizer
from model_compression_toolkit.pytorch.constants import KERNEL
from model_compression_toolkit.pytorch.quantizer.fake_quant_builder import power_of_two_quantization, \
    symmetric_quantization, uniform_quantization
from model_compression_toolkit.pytorch.quantizer.fake_quant_builder import power_of_two_quantization
from model_compression_toolkit.pytorch.reader.graph_builders import DummyPlaceHolder, ConstantHolder

"""
Division of Pytorch modules by how they should be quantized.
KERNEL_OPS: Layers that their coefficients should be quantized.
ACTIVATION: Layers that their activation should be quantized.
NO_QUANTIZATION: Layers that should not be quantized.
"""

KERNEL_OPS = [Conv2d, Linear, ConvTranspose2d]

NO_QUANTIZATION = [Dropout, Flatten, ConstantHolder] + \
                  [dropout, flatten, split, operator.getitem, reshape, unsqueeze]

ACTIVATION = [DummyPlaceHolder] + \
    [Hardswish, Hardsigmoid, ReLU, Hardtanh, ReLU6, LeakyReLU, PReLU, SiLU, Softmax, Sigmoid, Softplus, Softsign, Tanh] + \
    [hardswish, hardsigmoid, relu, hardtanh, relu6, leaky_relu, prelu, silu, softmax, sigmoid, softplus, softsign, tanh] + \
    [torch.relu] + \
    [UpsamplingBilinear2d, AdaptiveAvgPool2d, AvgPool2d, MaxPool2d] + \
    [upsample_bilinear, adaptive_avg_pool2d, avg_pool2d, max_pool2d] + \
    [add, sub, mul, multiply] + \
    [operator.add, operator.sub, operator.mul] + \
    [BatchNorm2d, concat, cat, mean]

"""
Map each layer to a list of its' weights attributes that should get quantized.
If a layer that is not listed here is queried, [None] is returned.
"""
KERNEL_ATTRIBUTES = DefaultDict({Conv2d: [KERNEL],
                                 ConvTranspose2d: [KERNEL],
                                 Linear: [KERNEL]},
                                lambda: [None])

"""
Map a layer to its kernel's output and input channels indices.
Map's values are tuples of (output_channel_index, input_channel_index).
Default value is returned for layers that are not included.
"""
DEFAULT_CHANNEL_AXIS_DICT = DefaultDict({Conv2d: (0, 1),
                                         Linear: (0, 1),
                                         ConvTranspose2d: (1, 0)},
                                        lambda: (None, None))

"""
Map from an activation function to its min/max output values (if known).
The values are used for tensor min/max values initialization.
"""
ACTIVATION2MINMAX = {} # should be an empty dict in Pytorch

"""
Map from an Pytorch module to its min/max output values (if known).
The values are used for tensor min/max values initialization.
"""
LAYER2MINMAX = {Softmax: (0, 1),
                softmax: (0, 1),
                Sigmoid: (0, 1),
                sigmoid: (0, 1),
                Hardsigmoid: (0, 1),
                hardsigmoid: (0, 1),
                ReLU: (0, None),
                relu: (0, None),
                ReLU6: (0, None),
                relu6: (0, None)}

"""
Mapping from a QuantizationMethod to an activation quantizer function.
"""
ACTIVATION_QUANTIZER_MAPPING = {QuantizationMethod.POWER_OF_TWO: power_of_two_quantization,
                                QuantizationMethod.SYMMETRIC: symmetric_quantization,
                                QuantizationMethod.UNIFORM: uniform_quantization}

"""
Mapping from a QuantizationMethod to an weights quantizer function.
"""
WEIGHTS_QUANTIZER_MAPPING = {QuantizationMethod.POWER_OF_TWO: power_of_two_quantizer,
                             QuantizationMethod.SYMMETRIC: symmetric_quantizer,
                             QuantizationMethod.UNIFORM: uniform_quantizer,
                             QuantizationMethod.KMEANS: kmeans_quantizer,
                             QuantizationMethod.LUT_QUANTIZER: lut_kmeans_quantizer}

"""
Output channel index of the model's layers
"""
OUTPUT_CHANNEL_INDEX = ChannelAxis.NCHW

DEFAULT_PYTORCH_INFO = FrameworkInfo(KERNEL_OPS,
                                     ACTIVATION,
                                     NO_QUANTIZATION,
                                     ACTIVATION_QUANTIZER_MAPPING,
                                     WEIGHTS_QUANTIZER_MAPPING,
                                     DEFAULT_CHANNEL_AXIS_DICT,
                                     ACTIVATION2MINMAX,
                                     LAYER2MINMAX,
                                     KERNEL_ATTRIBUTES,
                                     OUTPUT_CHANNEL_INDEX)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf
if tf.__version__ < "2.6":
    from tensorflow.keras.layers import Conv2D, Dense, Reshape, ZeroPadding2D, AveragePooling2D, Activation, DepthwiseConv2D, MaxPooling2D, ReLU, Add, Softmax, Concatenate, Multiply, Maximum, Minimum, BatchNormalization
else:
    from keras.layers import Conv2D, Dense, Reshape, ZeroPadding2D, AveragePooling2D, Activation, DepthwiseConv2D, MaxPooling2D, ReLU, Add, Softmax, Concatenate, Multiply, Maximum, Minimum, BatchNormalization


from tensorflow.python.keras.layers.core import SlicingOpLambda
from tensorflow.python.ops.image_ops_impl import ResizeMethod

from model_compression_toolkit.common.hardware_representation import FrameworkHardwareModel
from model_compression_toolkit.common.hardware_representation.hardware2framework import OperationsSetToLayers, \
    LayerFilterParams
from model_compression_toolkit.common.hardware_representation.hardware2framework.attribute_filter import Eq
from model_compression_toolkit.hardware_models.tflite import get_tflite_hw_model


def get_keras_hardware_model_tflite():
    tflite_hm = get_tflite_hw_model()
    tflite_keras = FrameworkHardwareModel(tflite_hm, name='tflite_keras')

    with tflite_keras:
        OperationsSetToLayers("PreserveQuantizationParams", [AveragePooling2D,
                                                             tf.nn.avg_pool2d,
                                                             Concatenate,
                                                             tf.concat,
                                                             MaxPooling2D,
                                                             Multiply,
                                                             tf.multiply,
                                                             Reshape,
                                                             tf.reshape,
                                                             LayerFilterParams(tf.image.resize,
                                                                               method=ResizeMethod.BILINEAR),
                                                             tf.nn.space_to_depth,
                                                             ZeroPadding2D,
                                                             tf.gather,
                                                             tf.compat.v1.batch_to_space_nd,
                                                             tf.space_to_batch_nd,
                                                             tf.transpose,
                                                             tf.maximum,
                                                             Maximum,
                                                             tf.minimum,
                                                             Minimum,
                                                             tf.pad,
                                                             tf.slice,
                                                             SlicingOpLambda])

        OperationsSetToLayers("FullyConnected", [Dense])
        OperationsSetToLayers("L2Normalization", [tf.math.l2_normalize])
        OperationsSetToLayers("LogSoftmax", [tf.nn.log_softmax])
        OperationsSetToLayers("Tanh", [tf.nn.tanh,
                                       LayerFilterParams(Activation, activation="tanh")])

        OperationsSetToLayers("Softmax", [tf.nn.softmax,
                                          Softmax,
                                          LayerFilterParams(Activation, activation="softmax")])

        OperationsSetToLayers("Logistic", [tf.sigmoid,
                                           LayerFilterParams(Activation, activation="sigmoid")])

        OperationsSetToLayers("Conv2d", [Conv2D])
        OperationsSetToLayers("DepthwiseConv2D", [DepthwiseConv2D])

        OperationsSetToLayers("Relu", [tf.nn.relu,
                                       tf.nn.relu6,
                                       LayerFilterParams(ReLU, Eq("max_value", None) | Eq("max_value", 6)),
                                       LayerFilterParams(Activation, activation="relu")])

        OperationsSetToLayers("Elu", [tf.nn.elu,
                                      LayerFilterParams(Activation, activation="elu")])

        OperationsSetToLayers("BatchNorm", [BatchNormalization,
                                            tf.nn.batch_normalization])

        OperationsSetToLayers("Squeeze", [tf.squeeze])
        OperationsSetToLayers("BiasAdd", [tf.nn.bias_add])
        OperationsSetToLayers("Add", [tf.add,
                                      Add])

    return tflite_keras




# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Any, List

import torch
import copy

from model_compression_toolkit import FrameworkInfo
from model_compression_toolkit.common import BaseNode
from model_compression_toolkit.pytorch.utils import set_model


class PytorchMixedPrecisionWrapper(torch.nn.Module):
    """
    Class that wraps a Pytorch layer (nn.Module) to be used for mixed precision quantization.
    Allows to maintain quantized weights tensors for each of the layer's attributes that we want to quantize,
    with each of the candidate bitwidth options specified for the mixed precision model.
    During MP search, it allows to activate the relevant quantized weights tensor according to a given configuration,
    and use it for inference.
    """
    def __init__(self,
                 n: BaseNode,
                 fw_info: FrameworkInfo):
        """
        Construct a Pytorch model that constitutes as a wrapper for a Pytorch layer, built from a given graph node.
        Args:
            n: Node to build its Pytorch layer.
            fw_info: Framework information (e.g., mapping from layers to their attributes to quantize).
        """
        super(PytorchMixedPrecisionWrapper, self).__init__()

        assert n.candidates_quantization_cfg is not None

        framework_attr = copy.copy(n.framework_attr)
        self.layer = n.type(**framework_attr)
        # loading the weights from the graph node (weights of the trained model)
        self.layer.load_state_dict({k: torch.Tensor(v) for k, v in n.weights.items()}, strict=False)
        set_model(self.layer)

        self.weight_attrs = fw_info.get_kernel_op_attributes(n.type)
        # float_weights is a list of weights for each attribute that we want to quantize.
        self.float_weights = [n.get_weights_by_keys(attr) for attr in
                              self.weight_attrs]

        assert len(self.weight_attrs) == len(self.float_weights)

        self.node_q_cfg = n.candidates_quantization_cfg
        self.quantizer_fn_list = [qc.weights_quantization_cfg.weights_quantization_fn for qc in self.node_q_cfg]
        self.quantized_weights = self._get_quantized_weights()

    def forward(self, x: Any) -> Any:
        """
        Args:
            x: input tensors to model.
        Returns:
            torch Tensor which is the output of the wrapped layer on the given input.
        """
        return self.layer(x)

    def _get_quantized_weights(self):
        """
        Calculates the quantized weights' tensors for each of the bitwidth candidates for quantization,
        to be stored and used during MP search.
        Returns: a list of quantized weights - for each bitwidth and layer's attribute to be quantized.
        """
        quantized_weights = []
        for index, qc in enumerate(self.node_q_cfg):
            # for each quantization configuration in mixed precision
            # get quantized weights for each attribute and for each filter
            quantized_per_attr = []
            for float_weight in self.float_weights:
                # for each attribute
                quantized_per_attr.append(self.quantizer_fn_list[index](tensor_data=float_weight,
                                                                        n_bits=qc.weights_quantization_cfg.weights_n_bits,
                                                                        signed=True,
                                                                        quantization_params=qc.weights_quantization_cfg.weights_quantization_params,
                                                                        per_channel=qc.weights_quantization_cfg.weights_per_channel_threshold,
                                                                        output_channels_axis=qc.weights_quantization_cfg.weights_channels_axis))
            quantized_weights.append(quantized_per_attr)

        return quantized_weights

    def set_active_weights(self,
                           bitwidth_idx: int,
                           attr: str = None):
        """
        Set a weights' tensor to use by the layer wrapped by the module.
        Args:
            bitwidth_idx: Index of a candidate quantization configuration to use its quantized
            version of the float weight.
            attr: Attributes of the layer's weights to quantize
        """
        if attr is None:  # set bit width to all weights of the layer
            attr_idxs = [attr_idx for attr_idx in range(len(self.quantized_weights[bitwidth_idx]))]
            self._set_bit_width_index(bitwidth_idx, attr_idxs)
        else:  # set bit width to a specific attribute
            attr_idx = self.weight_attrs.index(attr)
            self._set_bit_width_index(bitwidth_idx, [attr_idx])

    def _set_bit_width_index(self,
                             bitwidth_idx: int,
                             attr_idxs: List[int]):
        """
        Sets the wrapped layer's weights state with quantized weights, according to the given configuration.
        Args:
            bitwidth_idx: Index of a candidate quantization configuration to use its quantized
            version of the float weight.
            attr_idxs: Indices list of attributes of the layer's weights to quantize
        Returns: None (sets the new state of the layer inplace).
        """
        assert bitwidth_idx < len(self.quantized_weights), \
            f"Index {bitwidth_idx} does not exist in current quantization candidates list"

        loaded_weights = {k: torch.as_tensor(v) for k, v in self.layer.state_dict().items()}
        with torch.no_grad():
            for attr_idx in attr_idxs:
                # need to prepare the weights' tensor - extract it from the maintained quantized_weights list
                # and move it to the relevant device as the wrapped layer's weights.
                weights_tensor = self.quantized_weights[bitwidth_idx][attr_idx]
                weights_device = loaded_weights[self.weight_attrs[attr_idx]].device
                active_weights = torch.nn.Parameter(torch.from_numpy(weights_tensor).to(weights_device))
                loaded_weights[self.weight_attrs[attr_idx]] = active_weights
            self.layer.load_state_dict(loaded_weights, strict=True)



from typing import Dict, Any, Tuple, List

from model_compression_toolkit.common.graph.base_node import BaseNode
import numpy as np


class FunctionalNode(BaseNode):
    """
    Node that represents function ops with arguments to pass when building back the model.
    """

    def __init__(self,
                 name: str,
                 framework_attr: Dict[str, Any],
                 input_shape: Tuple[Any],
                 output_shape: Tuple[Any],
                 weights: Dict[str, np.ndarray],
                 layer_class: type,
                 op_call_args: List[Any] = None,
                 op_call_kwargs: Dict[str, Any] = None,
                 reuse: bool = False,
                 reuse_group: str = None,
                 quantization_attr: Dict[str, Any] = None,
                 functional_op: Any = None,
                 inputs_as_list: bool = False,
                 has_activation: bool = True):
        """
        Init a FunctionalNode object.

        Args:
            name: Node's name
            framework_attr: Framework attributes the layer had which the node holds.
            input_shape: Input tensor shape of the node.
            output_shape: Input tensor shape of the node.
            weights: Dictionary from a variable name to the weights with that name in the layer the node represents.
            layer_class: Class path of the layer this node represents.
            op_call_args: Arguments list to pass when calling the layer.
            op_call_kwargs: Key-Word Arguments dictionary with values to pass when calling the layer.
            reuse: Whether this node was duplicated and represents a reused layer.
            reuse_group: Name of group of nodes from the same reused layer.
            quantization_attr: Attributes the node holds regarding how it should be quantized.
            functional_op: The op the node implements.
            inputs_as_list: Whether to pass the node its input tensors as a list or not when calling the layer.
            has_activation: Whether the node has activations that we might want to quantize.

        """

        super().__init__(name,
                         framework_attr,
                         input_shape,
                         output_shape,
                         weights,
                         layer_class,
                         reuse,
                         reuse_group,
                         quantization_attr,
                         has_activation=has_activation)

        self.op_call_kwargs = op_call_kwargs
        self.op_call_args = op_call_args
        self.functional_op = functional_op
        self.inputs_as_list = inputs_as_list

    @property
    def type(self):
        """
        A function to get the node's function op for convenient comparison (instead of the layer_class)
        :return: the node's functional_op
        """
        return self.functional_op



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import logging
from typing import List, Dict, Tuple

import numpy as np
from tensorflow.keras.layers import DepthwiseConv2D, Conv2D, Dense, Conv2DTranspose, Activation, ReLU, ZeroPadding2D

from model_compression_toolkit import common
from model_compression_toolkit.common import Graph, BaseNode
from model_compression_toolkit.common.constants import OUTPUT_SCALE, THRESHOLD
from model_compression_toolkit.common.defaultdict import DefaultDict
from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.common.graph.graph_matchers import NodeOperationMatcher, WalkMatcher, \
    NodeFrameworkAttrMatcher
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig
from model_compression_toolkit.keras.constants import KERNEL, BIAS, LINEAR, ACTIVATION, RELU_MAX_VALUE
from model_compression_toolkit.keras.constants import RELU


# Match linear layers.
op2d_node = NodeOperationMatcher(DepthwiseConv2D) | \
            NodeOperationMatcher(Conv2D) | \
            NodeOperationMatcher(Conv2DTranspose) | \
            NodeOperationMatcher(Dense)

# Match Conv2D where its activation function keeps f(ax)==af(x)
homogeneous_activation_nodes = op2d_node & (NodeFrameworkAttrMatcher(ACTIVATION, RELU) |
                                            NodeFrameworkAttrMatcher(ACTIVATION, LINEAR))

zeropad_node = NodeOperationMatcher(ZeroPadding2D)

# The substitution is also possible for cases when there's a non-linearity
# between the two linear layers which keeps f(ax)==af(x)
mid_activation_nodes = (NodeOperationMatcher(Activation) &
                        NodeFrameworkAttrMatcher(ACTIVATION, RELU)) | \
                       NodeOperationMatcher(ReLU)

# Two cases to match: linear_op -> linear_op, linear_op -> non_linearity -> linear_op
# Two substitutions do the same thing but match different patterns.
MATCHER = WalkMatcher([homogeneous_activation_nodes, op2d_node])
MATCHER_WITH_PAD = WalkMatcher([homogeneous_activation_nodes, zeropad_node, op2d_node])
MATCHER_MID = WalkMatcher([op2d_node, mid_activation_nodes, op2d_node])
MATCHER_MID_WITH_PAD = WalkMatcher([op2d_node, mid_activation_nodes, zeropad_node, op2d_node])


def scale_reshaping(scale: np.ndarray,
                    op2d: common.BaseNode,
                    kernel_channel_mapping: DefaultDict,
                    in_channels: bool = True) -> np.ndarray:
    """
    Before scaling a kernel, the scale factor needs is reshaped to the correct
    dimensions. This is a function of the layer that is scaled and whether its input channels or
    output channels that should be scaled.
    The index of the correct kernel index is obtained from kernel_channel_mapping.

    Args:
        scale: Scale factor to scale the kernel channels by.
        op2d: Node to scale its kernel.
        kernel_channel_mapping: Mapping from a layer to a tuple of indices of its output/input kernel channels.
        in_channels: Kernel's index of input channels.

    Returns:
        The scale factor after reshaping it to the correct shape.
    """

    op_ndims = op2d.get_weights_by_keys(KERNEL).ndim
    reshape_target = np.ones(op_ndims, dtype=np.int)
    reshape_target[kernel_channel_mapping.get(op2d.type)[int(in_channels)]] = -1
    return np.reshape(scale, reshape_target)


def update_linear_nodes(graph:Graph,
                        qc: QuantizationConfig,
                        fw_info: FrameworkInfo,
                        first_op2d_node: BaseNode,
                        second_op2d_node: BaseNode,
                        scale_factor: np.ndarray):
    """
    Scale the weights of two linear nodes with a scale factor. Each node is scaled in
    the opposite scale factor such that the output of the second node is the same as it
    is without the scaling.
    Thresholds are recalculated as the weights were changed.
    The scale factor contain a scale value per-channel.

    Args:
        graph: Graph to apply the scaling on its nodes.
        qc: QuantizationConfig containing parameters of how the model should be quantized.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
        groups of layers by how they should be quantized, etc.)
        first_op2d_node: Node to multiply its kernel by the scale factor.
        second_op2d_node: Node to divide its kernel by the scale factor.
        scale_factor: Scaling factor to scale the nodes' weights.

    """

    w2_fixed = second_op2d_node.get_weights_by_keys(KERNEL) / scale_reshaping(scale_factor,
                                                                              second_op2d_node,
                                                                              fw_info.kernel_channels_mapping)

    w1_fixed = first_op2d_node.get_weights_by_keys(KERNEL) * scale_reshaping(scale_factor,
                                                                             first_op2d_node,
                                                                             fw_info.kernel_channels_mapping,
                                                                             in_channels=False)

    if first_op2d_node.get_weights_by_keys(BIAS) is not None:
        b1_fixed = first_op2d_node.get_weights_by_keys(BIAS) * scale_factor
        first_op2d_node.set_weights_by_keys(BIAS, b1_fixed)

    first_op2d_node.quantization_attr[OUTPUT_SCALE] = scale_factor
    first_op2d_node.set_weights_by_keys(KERNEL, w1_fixed)
    second_op2d_node.set_weights_by_keys(KERNEL, w2_fixed)

    for nqc in first_op2d_node.candidates_quantization_cfg:
        nqc.weights_quantization_cfg.calculate_and_set_weights_params(w1_fixed)
    for nqc in second_op2d_node.candidates_quantization_cfg:
        nqc.weights_quantization_cfg.calculate_and_set_weights_params(w2_fixed)


def calculate_scale_correction(graph: Graph,
                               activation_node: BaseNode,
                               eps: float = 1e-6) -> tuple:
    """
    Compute a scale factor by the activation node threshold and its outputs statistics in
    order to scale all activations such that their maximal values are the activation node's
    constrained threshold.

    Args:
        graph: Graph where the activation node is in.
        activation_node: Activation node to scale its outputs (thus, its previous layer weights).
        eps: Small number to use the the maximal statistics values are zero.

    Returns:
        Tuple of: scaling factor, activation node constrained threshold, outputs maximal value per-channel.
    """

    tensor_stat = graph.get_out_stats_collector(activation_node)

    if not activation_node.is_all_activation_candidates_equal():
        raise Exception("Scale equalization is not supported for more than one activation quantization configuration "
                        "candidate")

    # all candidates have same activation config, so taking the first candidate for calculations
    threshold = activation_node.candidates_quantization_cfg[0].activation_quantization_cfg.activation_quantization_params.get(THRESHOLD)

    max_vector = np.max(np.stack([tensor_stat.mpcc.max_per_channel, np.abs(tensor_stat.mpcc.min_per_channel)], axis=-1),
                        axis=-1)

    scale_factor = threshold / (max_vector + eps)
    scale_factor[max_vector <= 0] = 1
    activation_node.quantization_attr[OUTPUT_SCALE] = scale_factor
    scale_factor = np.maximum(scale_factor, 1)  # Making sure all scale factor are above 1

    graph.scale_stats_collector(activation_node, scale_factor)

    # scale relu bound so f(ax)==af(x)
    if activation_node.type == ReLU and \
            activation_node.framework_attr.get(RELU_MAX_VALUE) is not None:
        activation_node.framework_attr[RELU_MAX_VALUE] = threshold

    return scale_factor, threshold, tensor_stat.mpcc.state


def scale_equalization_lnl(graph: Graph,
                           qc: QuantizationConfig,
                           fw_info: FrameworkInfo,
                           first_op2d_node: BaseNode,
                           n_node: BaseNode,
                           second_op2d_node: BaseNode):
    """
    Compute a scale factor to scale all activation node's outputs such that
    its maximum per-channel is the constrained threshold of the activation node.
    A correction (opposite computed scale) needs to be applied on the linear node that
    follows the activation node to get the same expected output without the scaling.

    Args:
        graph: Graph to apply the scaling on its nodes.
        n_node: Activation node in the middle of the linear nodes.
        qc: QuantizationConfig containing parameters of how the model should be quantized.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
        groups of layers by how they should be quantized, etc.)
        first_op2d_node: Node to multiply its kernel by the scale factor.
        second_op2d_node: Node to divide its kernel by the scale factor.

    """
    scale_factor, threshold, max_array = calculate_scale_correction(graph, n_node)

    common.Logger.debug(f"{first_op2d_node.name} -> output Max Per Channel:{max_array}")
    common.Logger.debug(f'{first_op2d_node.name} -> Threshold value: {threshold}')
    common.Logger.debug(f'{first_op2d_node.name} -> Scale Factor: {scale_factor}')

    update_linear_nodes(graph,
                        qc,
                        fw_info,
                        first_op2d_node,
                        second_op2d_node,
                        scale_factor)


class BaseScaleEqualization(common.BaseSubstitution):
    """
    Substitution to scale the weights of two linear nodes in order to use the entire
    constrained range when activations are quantized.
    Unlike relu_bound_to_power_of_2, scaling here is per-channel and a non-linear node my or may not
    exist between the linear nodes (set nl_index to 0 if it doesn't)
    """

    def __init__(self,
                 quant_config: QuantizationConfig,
                 fw_info: FrameworkInfo,
                 matcher_instance,
                 nl_index: int = 0):
        """
        Initialize a ScaleEqualization object.
        Args:
            quant_config: QuantizationConfig containing parameters of how the model should be quantized.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
            groups of layers by how they should be quantized, etc.)
            matcher_instance: Per substitution matcher instance of type WalkMatcher
            nl_index: Index of non-linear node in list of matched nodes (index 0 means no non-linear in matcher)
        """

        self.quant_config = quant_config
        self.fw_info = fw_info
        self.nl_index = nl_index
        super().__init__(matcher_instance=matcher_instance)

    def substitute(self,
                   graph: Graph,
                   nodes_list: List[BaseNode]) -> Graph:
        """
        Scale each channel of the weights of two linear nodes,
        in order to use the entire constrained range when activations are quantized.
        If the non-linear layer is bounded, the bound is scaled as well.

        Args:
            graph: Graph to apply the substitution on.
            nodes_list: List of nodes that match the pattern in the substitution init.

        Returns:
            Graph after substitution.

        """
        first_op2d_node = nodes_list[0]
        nl_node = nodes_list[self.nl_index]
        second_op2d_node = nodes_list[-1]
        scale_equalization_lnl(graph,
                               self.quant_config,
                               self.fw_info,
                               first_op2d_node,
                               nl_node,
                               second_op2d_node)
        return graph


class ScaleEqualization(BaseScaleEqualization):
    """
    Substitution extends BaseScaleEqualization to the case of Linear-->Linear
    """

    def __init__(self,
                 quant_config: QuantizationConfig,
                 fw_info: FrameworkInfo):
        """
        Initialize a ScaleEqualization object.
        Args:
            quant_config: QuantizationConfig containing parameters of how the model should be quantized.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
            groups of layers by how they should be quantized, etc.)
        """

        super().__init__(quant_config=quant_config, fw_info=fw_info, matcher_instance=MATCHER)


class ScaleEqualizationWithPad(BaseScaleEqualization):
    """
    Substitution extends BaseScaleEqualization to the case of Linear-->ZeroPadding-->Linear
    """

    def __init__(self,
                 quant_config: QuantizationConfig,
                 fw_info: FrameworkInfo):
        """
        Initialize a ScaleEqualization object.
        Args:
            quant_config: QuantizationConfig containing parameters of how the model should be quantized.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
            groups of layers by how they should be quantized, etc.)
        """

        super().__init__(quant_config=quant_config, fw_info=fw_info, matcher_instance=MATCHER_WITH_PAD)


class ScaleEqualizationMidActivation(BaseScaleEqualization):
    """
    Substitution extends BaseScaleEqualization to the case of Linear-->Non-Lnear-->Linear
    """

    def __init__(self,
                 quant_config: QuantizationConfig,
                 fw_info: FrameworkInfo):
        """
        Init a ScaleEqualizationMidActivation object.

        Args:
            quant_config: QuantizationConfig containing parameters of how the model should be quantized.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
            groups of layers by how they should be quantized, etc.)

        """

        self.quant_config = quant_config
        self.fw_info = fw_info

        super().__init__(quant_config=quant_config, fw_info=fw_info, matcher_instance=MATCHER_MID, nl_index=1)


class ScaleEqualizationMidActivationWithPad(BaseScaleEqualization):
    """
    Substitution extends BaseScaleEqualization to the case of Linear-->Non-linear-->ZeroPadding-->Linear
    """

    def __init__(self,
                 quant_config: QuantizationConfig,
                 fw_info: FrameworkInfo):
        """
        Init a ScaleEqualizationMidActivation object.

        Args:
            quant_config: QuantizationConfig containing parameters of how the model should be quantized.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
            groups of layers by how they should be quantized, etc.)

        """

        self.quant_config = quant_config
        self.fw_info = fw_info

        super().__init__(quant_config=quant_config, fw_info=fw_info, matcher_instance=MATCHER_MID_WITH_PAD, nl_index=1)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import inspect
from typing import Dict, List, Tuple, Callable
import torch
from torch.fx import GraphModule

from model_compression_toolkit import common
from model_compression_toolkit.common import BaseNode
from model_compression_toolkit.common.graph.base_graph import OutTensor
from model_compression_toolkit.common.graph.edge import Edge
from model_compression_toolkit.common.graph.functional_node import FunctionalNode
from model_compression_toolkit.pytorch.constants import OUTPUT, PLACEHOLDER, TENSOR_META, CALL_FUNCTION, TYPE, \
    CALL_METHOD, BIAS, FUNCTIONAL_OP, OP_CALL_KWARGS, OP_CALL_ARGS, INPUTS_AS_LIST, GET_ATTR, CONSTANT


class DummyPlaceHolder(torch.nn.Module):
    """
    Class for PlaceHolder operator since a Pytorch model doesn't have one but FX does.
    """
    def __name__(self):
        return PLACEHOLDER

    def forward(self, x):
        return x


class ConstantHolder(torch.nn.Module):
    """
    Class for saving constant values or parameters in graph inference.
    """
    def __init__(self, const_size):
        super(ConstantHolder, self).__init__()
        setattr(self, CONSTANT, torch.nn.Parameter(torch.empty(const_size)))

    def __name__(self):
        return CONSTANT

    def forward(self):
        return getattr(self, CONSTANT)


def nodes_builder(model: GraphModule,
                  module_dict: Dict,
                  to_numpy: Callable) -> Tuple[List, List, List, Dict]:
    """
    Build a node from a fx node. A node contains all information to reconstruct the model module or call function
    it's representing in the model: operation, module configuration, weights, input/output shape.
    Args:
        model: Pytorch FX model.
        module_dict: A dictionary of the Pyotrch model's named modules.

    Returns:
        A list of Graph nodes that were built from the fx GraphModule nodes.
    """
    # init function variables:
    inputs = []
    outputs = []
    nodes = []
    output_nodes = []
    fx_node_2_graph_node = {}

    for node in model.graph.nodes:
        # extract node type and framework attributes
        framework_attr = dict(node.kwargs)
        node_has_activation = True
        if node.target in module_dict.keys():
            node_module = module_dict[node.target]
            node_type = type(node_module)
            framework_attr = node_module.__dict__
            fullargspec = inspect.getfullargspec(node_type.__init__).args
            framework_attr = {k: v for k, v in framework_attr.items() if k in fullargspec}
            if hasattr(node_module, BIAS) and BIAS in fullargspec:
                framework_attr[BIAS] = False if node_module.bias is None else True
        elif node.op == CALL_FUNCTION:
            node_type = node.target
        elif node.op == PLACEHOLDER:
            node_type = DummyPlaceHolder
        elif node.op == OUTPUT:
            output_nodes += node.all_input_nodes
            continue
        elif node.op == CALL_METHOD:
            if hasattr(torch, node.target):
                node_type = getattr(torch, node.target)
            elif hasattr(torch.Tensor, node.target):
                node_type = getattr(torch.Tensor, node.target)
            else:
                raise Exception(f'Call method of type \'{node.target}\' is currently not supported.')
        elif node.op == GET_ATTR:
            node_type = ConstantHolder
            node_has_activation = False
            common.Logger.warning(
                'Pytorch model has a parameter or constant Tensor value. This can cause unexpected behaviour when '
                'converting the model.')
        else:
            raise Exception(f'Unknown node type: {node.name}')

        # extract layer weights and named buffers
        weights = {}
        if node.target in module_dict.keys():
            named_parameters_weights = {name: to_numpy(parameter) for name, parameter in
                                        module_dict[node.target].named_parameters()}
            named_buffer_weights = {name: to_numpy(parameter) for name, parameter in
                                    module_dict[node.target].named_buffers() if len(parameter.shape) > 0}
            weights.update(named_parameters_weights)
            weights.update(named_buffer_weights)

        if node.op == GET_ATTR:
            named_parameters_weights = {CONSTANT: to_numpy(parameter) for name, parameter in
                                        model.named_parameters() if node.target == name}
            named_buffer_weights = {CONSTANT: to_numpy(parameter) for name, parameter in
                                    model.named_buffers() if node.target == name}
            if len(named_parameters_weights) + len(named_buffer_weights) > 1:
                raise Exception(
                    f'Constant parameter can only have one tensor. Here we have {len(named_parameters_weights) + len(named_buffer_weights)}')

            weights.update(named_parameters_weights)
            weights.update(named_buffer_weights)
            framework_attr.update(const_size=weights.get(CONSTANT).shape)

        # extract input shapes
        input_shape = []
        if node.op != PLACEHOLDER:
            for input_node in node.all_input_nodes:
                tensor_meta = input_node.meta
                if tensor_meta[TYPE] == torch.Tensor:
                    input_shape += [list(tensor_meta[TENSOR_META].shape)]
                elif tensor_meta[TYPE] == tuple:
                    input_shape += [list(n.shape) for n in tensor_meta[TENSOR_META]]

        # extract output shapes
        if node.meta[TYPE] == torch.Tensor:
            output_shape = [list(node.meta[TENSOR_META].shape)]
        elif node.meta[TYPE] in (list, tuple):
            output_shape = [list(m.shape) for m in node.meta[TENSOR_META]]
        else:
            output_shape = []

        # initiate graph nodes
        if node.op in [CALL_METHOD, CALL_FUNCTION]:
            graph_node_type = FunctionalNode
            inputs_as_list1 = len(node.args) > 0 and isinstance(node.args[0], (list, tuple)) and all(
                [isinstance(n, torch.fx.node.Node) for n in node.args[0]])
            inputs_as_list = inputs_as_list1 or \
                             (len(node.args) > 0 and node.args[0].op == PLACEHOLDER and node.args[0].meta[TYPE] in (list, tuple))
            if inputs_as_list:
                num_inputs = 1
            else:
                input_counter = 0
                for in_node in node.all_input_nodes:
                    for arg in node.args:
                        if arg == in_node:
                            input_counter += 1
                num_inputs = max(len(node.all_input_nodes), input_counter)
            op_call_args = list(node.args[num_inputs:])
            kwargs = {FUNCTIONAL_OP: node_type,
                      OP_CALL_ARGS: op_call_args,
                      OP_CALL_KWARGS: node.kwargs,
                      INPUTS_AS_LIST: inputs_as_list}
        else:
            graph_node_type = BaseNode
            kwargs = {}
        graph_node = graph_node_type(name=node.name,
                                     framework_attr=framework_attr,
                                     input_shape=input_shape,
                                     output_shape=output_shape,
                                     weights=weights,
                                     layer_class=node_type,
                                     has_activation=node_has_activation,
                                     **kwargs)

        # generate graph inputs list
        if node.op == PLACEHOLDER:
            for ii in range(len(output_shape)):
                inputs.append(graph_node)

        fx_node_2_graph_node[node] = graph_node
        nodes.append(graph_node)

    # generate graph outputs list
    for node in output_nodes:
        outputs.append(OutTensor(fx_node_2_graph_node[node], output_nodes.index(node)))

    return nodes, inputs, outputs, fx_node_2_graph_node


def edges_builder(model: GraphModule,
                   fx_node_2_graph_node: Dict) -> List:
    """

    Args:
        model: Pytorch FX model.
        fx_node_2_graph_node: dictionary from fx node to graph node.

    Returns:
        List of graph edges
    """
    src_index = 0 # in fx src_index is always zero because fx uses the getitem operator to fetch node outputs
    edges = []
    connectivity_dict = {}
    for node in model.graph.nodes:
        if node.op != OUTPUT:
            for input_node in node.all_input_nodes:

                # n_edges_for_input_node is for the case that the input node appears more than
                # once as the input of the node, for example add(x, x)
                n_edges_for_input_node = sum([1 for a in node.args if input_node == a])
                n_edges_for_input_node = max(n_edges_for_input_node, 1)

                dst_index = node.all_input_nodes.index(input_node)
                for i in range(n_edges_for_input_node):
                    if connectivity_dict.get(input_node):
                        connectivity_dict[input_node].append((node, dst_index))
                    else:
                        connectivity_dict[input_node] = [(node, dst_index)]
                    dst_index += 1
    for node in model.graph.nodes:
        out_nodes = connectivity_dict.get(node)
        if out_nodes:
            for (out_node, dst_index) in out_nodes:
                edges.append(
                    Edge(fx_node_2_graph_node[node], fx_node_2_graph_node[out_node], src_index, dst_index))

    return edges